# -*- coding: utf-8 -*-


"""
TODO: enable code debug and error report; See https://github.com/numba/numba/issues/7370
"""

import ast
import inspect
import logging
import re
from copy import deepcopy
from pprint import pprint

import numba
import numba.misc.help.inspector as inspector
import numpy as np
from numba.core.dispatcher import Dispatcher

from brainpy import errors, math, tools
from brainpy.base.base import Base
from brainpy.base.collector import Collector
from brainpy.base.function import Function
from brainpy.math import profile

DE_INT = DynamicalSystem = None
logger = logging.getLogger('brainpy.math.numpy.ast2numba')

__all__ = [
  'jit',
]


def jit(obj_or_fun, show_code=False, **jit_setting):
  global DE_INT
  if DE_INT is None:
    from brainpy.integrators.constants import DE_INT

  if callable(obj_or_fun):
    # Function
    if isinstance(obj_or_fun, Function):
      return jit_Func(obj_or_fun,
                      show_code=show_code,
                      **jit_setting)

    # Base
    elif isinstance(obj_or_fun, Base):
      return jit_Base(func=obj_or_fun.__call__,
                      host=obj_or_fun,
                      name=obj_or_fun.name + '_call',
                      show_code=show_code, **jit_setting)

    # integrator
    elif hasattr(obj_or_fun, '__name__') and obj_or_fun.__name__.startswith(DE_INT):
      return jit_integrator(intg=obj_or_fun,
                            show_code=show_code,
                            **jit_setting)

    # bounded method
    elif hasattr(obj_or_fun, '__self__') and isinstance(obj_or_fun.__self__, Base):
      return jit_Base(func=obj_or_fun,
                      host=obj_or_fun.__self__,
                      show_code=show_code,
                      **jit_setting)

    else:
      # native function
      if not isinstance(obj_or_fun, Dispatcher):
        return numba.jit(obj_or_fun, **jit_setting)
      else:
        # numba function
        return obj_or_fun

  else:
    return jit_DS(obj_or_fun,
                  show_code=show_code,
                  **jit_setting)


def jit_DS(obj_or_fun, show_code=False, **jit_setting):
  global DynamicalSystem
  if DynamicalSystem is None:
    from brainpy.simulation.brainobjects.base import DynamicalSystem

  if not isinstance(obj_or_fun, DynamicalSystem):
    raise errors.UnsupportedError(f'JIT compilation in numpy backend only '
                                  f'supports {Base.__name__}, but we got '
                                  f'{type(obj_or_fun)}.')
  if not hasattr(obj_or_fun, 'steps'):
    raise errors.BrainPyError(f'Please init this DynamicalSystem {obj_or_fun} first, '
                              f'then apply JIT.')

  # function analysis
  for key, step in list(obj_or_fun.steps.items()):
    key = key.replace(".", "_")
    r = _jit_func(obj_or_fun=step, show_code=show_code, **jit_setting)
    if r['func'] != step:
      func = _form_final_call(f_org=step, f_rep=r['func'], arg2call=r['arg2call'],
                              arguments=r['arguments'], nodes=r['nodes'],
                              show_code=show_code, name=step.__name__)
      obj_or_fun.steps.replace(key, func)

  # dynamic system
  return obj_or_fun


def jit_integrator(intg, show_code=False, **jit_setting):
  r = _jit_intg_func(intg, show_code=show_code, **jit_setting)
  if len(r['arguments']):
    intg = _form_final_call(f_org=intg, f_rep=r['func'], arg2call=r['arg2call'],
                            arguments=r['arguments'], nodes=r['nodes'],
                            show_code=show_code, name=intg.__name__)
  else:
    intg = r['func']
  return intg


def jit_Func(func, show_code=False, **jit_setting):
  assert isinstance(func, Function)

  r = _jit_Function(func=func, show_code=show_code, **jit_setting)
  if len(r['arguments']):
    func = _form_final_call(f_org=func._f, f_rep=r['func'], arg2call=r['arg2call'],
                            arguments=r['arguments'], nodes=r['nodes'],
                            show_code=show_code, name=func.name + '_call')
  else:
    func = r['func']

  return func


def jit_Base(func, host, name=None, show_code=False, **jit_setting):
  r = _jit_cls_func(func, host=host, show_code=show_code, **jit_setting)
  if len(r['arguments']):
    name = func.__name__ if name is None else name
    func = _form_final_call(f_org=func, f_rep=r['func'], arg2call=r['arg2call'],
                            arguments=r['arguments'], nodes=r['nodes'],
                            show_code=show_code, name=name)
  else:
    func = r['func']
  return func


def _jit_func(obj_or_fun, show_code=False, **jit_setting):
  global DE_INT
  if DE_INT is None:
    from brainpy.integrators.constants import DE_INT

  if callable(obj_or_fun):
    # integrator
    if hasattr(obj_or_fun, '__name__') and obj_or_fun.__name__.startswith(DE_INT):
      return _jit_intg_func(obj_or_fun,
                            show_code=show_code,
                            **jit_setting)

    # bounded method
    elif hasattr(obj_or_fun, '__self__') and isinstance(obj_or_fun.__self__, Base):
      return _jit_cls_func(obj_or_fun,
                           host=obj_or_fun.__self__,
                           show_code=show_code,
                           **jit_setting)

    # wrapped function
    elif isinstance(obj_or_fun, Function):
      return _jit_Function(obj_or_fun,
                           show_code=show_code,
                           **jit_setting)

    # base class function
    elif isinstance(obj_or_fun, Base):
      return _jit_cls_func(obj_or_fun.__call__,
                           host=obj_or_fun,
                           show_code=show_code,
                           **jit_setting)

    else:
      # native function
      if not isinstance(obj_or_fun, Dispatcher):
        if inspector.inspect_function(obj_or_fun)['numba_type'] is None:
          f = numba.jit(obj_or_fun, **jit_setting)
          return dict(func=f, arguments=set(), arg2call=Collector(), nodes=Collector())
      # numba function or innate supported function
      return dict(func=obj_or_fun, arguments=set(), arg2call=Collector(), nodes=Collector())

  else:
    raise ValueError


def _jit_Function(func, show_code=False, **jit_setting):
  assert isinstance(func, Function)

  # code_scope
  closure_vars = inspect.getclosurevars(func._f)
  code_scope = dict(closure_vars.nonlocals)
  code_scope.update(closure_vars.globals)
  # code
  code = tools.deindent(inspect.getsource(func._f)).strip()
  # arguments
  arguments = set()
  # nodes
  nodes = {v.name: v for v in func._nodes.values()}
  # arg2call
  arg2call = dict()

  for key, node in func._nodes.items():
    code, _arguments, _arg2call, _nodes, code_scope = _analyze_cls_func(
      host=node, code=code, show_code=show_code, code_scope=code_scope,
      self_name=key, pop_self=True, **jit_setting)
    arguments.update(_arguments)
    arg2call.update(_arg2call)
    nodes.update(_nodes)

  # compile new function
  # code, _scope = _add_try_except(code)
  # code_scope.update(_scope)
  if show_code:
    show_compiled_codes(code, code_scope)
  exec(compile(code, '', 'exec'), code_scope)
  func = code_scope[func._f.__name__]
  func = numba.jit(func, **jit_setting)

  # returns
  return dict(func=func, arguments=arguments, arg2call=arg2call, nodes=nodes)


def _jit_cls_func(f, code=None, host=None, show_code=False, **jit_setting):
  host = (host or f.__self__)

  # data to return
  arguments = set()
  arg2call = dict()
  nodes = Collector()
  nodes[host.name] = host

  # code
  code = (code or tools.deindent(inspect.getsource(f)).strip())
  # function name
  func_name = f.__name__
  # code scope
  closure_vars = inspect.getclosurevars(f)
  code_scope = dict(closure_vars.nonlocals)
  code_scope.update(closure_vars.globals)
  # analyze class function
  code, _arguments, _arg2call, _nodes, _code_scope = _analyze_cls_func(
    host=host, code=code, show_code=show_code, **jit_setting)
  arguments.update(_arguments)
  arg2call.update(_arg2call)
  nodes.update(_nodes)
  code_scope.update(_code_scope)

  # compile new function
  # code, _scope = _add_try_except(code)
  # code_scope.update(_scope)
  if show_code:
    show_compiled_codes(code, code_scope)
  exec(compile(code, '', 'exec'), code_scope)
  func = code_scope[func_name]
  func = numba.jit(func, **jit_setting)

  # returns
  return dict(func=func, arguments=arguments, arg2call=arg2call, nodes=nodes)


def _jit_intg_func(f, show_code=False, **jit_setting):
  global DynamicalSystem
  if DynamicalSystem is None:
    from brainpy.simulation.brainobjects.base import DynamicalSystem

  # exponential euler methods
  if f.brainpy_data['method'].startswith('exponential'):
    return _jit_cls_func(f=f, code="\n".join(f.brainpy_data['code_lines']),
                         show_code=show_code, **jit_setting)

  # information in the integrator
  func_name = f.brainpy_data['func_name']
  raw_func = f.brainpy_data['raw_func']
  tree = ast.parse('\n'.join(f.brainpy_data['code_lines']))
  code_scope = {key: val for key, val in f.brainpy_data['code_scope'].items()}

  # essential information
  arguments = set()
  arg2call = dict()
  nodes = Collector()

  # jit raw functions
  f_node = None
  remove_self = None
  if hasattr(f, '__self__') and isinstance(f.__self__, DynamicalSystem):
    f_node = f.__self__
    _arg = tree.body[0].args.args.pop(0)  # remove "self" arg
    # remove "self" in functional call
    remove_self = _arg.arg

  need_recompile = False
  for key, func in raw_func.items():
    # get node of host
    func_node = None
    if f_node:
      func_node = f_node
    elif hasattr(func, '__self__') and isinstance(func.__self__, DynamicalSystem):
      func_node = func.__self__

    # get new compiled function
    if isinstance(func, Dispatcher):
      continue
    elif func_node is not None:
      need_recompile = True
      r = _jit_cls_func(f=func,
                        host=func_node,
                        show_code=show_code,
                        **jit_setting)
      if len(r['arguments']) or remove_self:
        tree = _replace_func_call_by_tree(tree,
                                          func_call=key,
                                          arg_to_append=r['arguments'],
                                          remove_self=remove_self)
      code_scope[key] = r['func']
      arguments.update(r['arguments'])  # update arguments
      arg2call.update(r['arg2call'])  # update arg2call
      nodes.update(r['nodes'])  # update nodes
      nodes[func_node.name] = func_node  # update nodes
    else:
      need_recompile = True
      code_scope[key] = numba.jit(func, **jit_setting)

  if need_recompile:
    tree.body[0].decorator_list.clear()
    tree.body[0].args.args.extend([ast.Name(id=a) for a in sorted(arguments)])
    tree.body[0].args.defaults.extend([ast.Constant(None) for _ in sorted(arguments)])
    code = tools.ast2code(tree)
    # code, _scope = _add_try_except(code)
    # code_scope.update(_scope)
    code_scope_backup = {k: v for k, v in code_scope.items()}
    # compile functions
    if show_code:
      show_compiled_codes(code, code_scope)
    exec(compile(code, '', 'exec'), code_scope)
    new_f = code_scope[func_name]
    new_f.brainpy_data = {key: val for key, val in f.brainpy_data.items()}
    new_f.brainpy_data['code_lines'] = code.strip().split('\n')
    new_f.brainpy_data['code_scope'] = code_scope_backup
    jit_f = numba.jit(new_f, **jit_setting)
    return dict(func=jit_f, arguments=arguments, arg2call=arg2call, nodes=nodes)
  else:
    return dict(func=f, arguments=arguments, arg2call=arg2call, nodes=nodes)


def _analyze_cls_func(host, code, show_code, self_name=None, pop_self=True, **jit_setting):
  """Analyze the bounded function of one object.

  Parameters
  ----------
  host : Base
    The data host.
  code : str
    The function source code.
  self_name : optional, str
    The class name, like "self", "cls".
  show_code : bool
  """
  # arguments
  tree = ast.parse(code)
  if self_name is None:
    self_name = tree.body[0].args.args[0].arg
    # data assigned by self.xx in line right
    if self_name not in profile.CLASS_KEYWORDS:
      raise errors.CodeError(f'BrainPy only support class keyword '
                             f'{profile.CLASS_KEYWORDS}, but we got {self_name}.')
  if pop_self:
    tree.body[0].args.args.pop(0)  # remove "self" etc. class argument

  # analyze function body
  r = _analyze_cls_func_body(host=host, self_name=self_name, code=code, tree=tree,
                             show_code=show_code, has_func_def=True, **jit_setting)
  code, arguments, arg2call, nodes, code_scope = r

  return code, arguments, arg2call, nodes, code_scope


def _analyze_cls_func_body(host, self_name, code, tree, show_code=False,
                           has_func_def=False, **jit_setting):
  arguments, arg2call, nodes, code_scope = set(), dict(), Collector(), dict()

  # all self data
  self_data = re.findall('\\b' + self_name + '\\.[A-Za-z_][A-Za-z0-9_.]*\\b', code)
  self_data = list(set(self_data))

  # analyze variables and functions accessed by the self.xx
  data_to_replace = {}
  for key in self_data:
    split_keys = key.split('.')
    if len(split_keys) < 2:
      raise errors.BrainPyError

    # get target and data
    target = host
    for i in range(1, len(split_keys)):
      next_target = getattr(target, split_keys[i])
      if not isinstance(next_target, Base):
        break
      target = next_target
    else:
      raise errors.BrainPyError
    data = getattr(target, split_keys[i])

    # analyze data
    if isinstance(data, math.Variable):  # data is a variable
      arguments.add(f'{target.name}_{split_keys[i]}')
      arg2call[f'{target.name}_{split_keys[i]}'] = f'{target.name}.{split_keys[-1]}.value'
      nodes[target.name] = target
      # replace the data
      if len(split_keys) == i + 1:
        data_to_replace[key] = f'{target.name}_{split_keys[i]}'
      else:
        data_to_replace[key] = f'{target.name}_{split_keys[i]}.{".".join(split_keys[i + 1:])}'

    elif isinstance(data, np.random.RandomState):  # data is a RandomState
      # replace RandomState
      code_scope[f'{target.name}_{split_keys[i]}'] = np.random
      # replace the data
      if len(split_keys) == i + 1:
        data_to_replace[key] = f'{target.name}_{split_keys[i]}'
      else:
        data_to_replace[key] = f'{target.name}_{split_keys[i]}.{".".join(split_keys[i + 1:])}'

    elif callable(data):  # data is a function
      assert len(split_keys) == i + 1
      r = _jit_func(obj_or_fun=data, show_code=show_code, **jit_setting)
      # if len(r['arguments']):
      tree = _replace_func_call_by_tree(tree, func_call=key, arg_to_append=r['arguments'])
      arguments.update(r['arguments'])
      arg2call.update(r['arg2call'])
      nodes.update(r['nodes'])
      code_scope[f'{target.name}_{split_keys[i]}'] = r['func']
      data_to_replace[key] = f'{target.name}_{split_keys[i]}'  # replace the data

    elif isinstance(data, (dict, list, tuple)):  # data is a list/tuple/dict of function/object
      # get all values
      if isinstance(data, dict):  # check dict
        if len(split_keys) != i + 2 and split_keys[-1] != 'values':
          raise errors.BrainPyError(f'Only support iter dict.values(). while we got '
                                    f'dict.{split_keys[-1]}  for data: \n\n{data}')
        values = list(data.values())
        iter_name = key + '()'
      else:  # check list / tuple
        assert len(split_keys) == i + 1
        values = list(data)
        iter_name = key
        if len(values) > 0:
          if not (callable(values[0]) or isinstance(values[0], Base)):
            code_scope[f'{target.name}_{split_keys[i]}'] = data
            if len(split_keys) == i + 1:
              data_to_replace[key] = f'{target.name}_{split_keys[i]}'
            else:
              data_to_replace[key] = f'{target.name}_{split_keys[i]}.{".".join(split_keys[i + 1:])}'
            continue
            raise errors.BrainPyError(f'Only support JIT an iterable objects of function '
                                      f'or Base object, but we got:\n\n {values[0]}')
      # replace this for-loop
      r = _replace_this_forloop(tree=tree,
                                iter_name=iter_name,
                                loop_values=values,
                                show_code=show_code,
                                **jit_setting)
      tree, _arguments, _arg2call, _nodes, _code_scope = r
      arguments.update(_arguments)
      arg2call.update(_arg2call)
      nodes.update(_nodes)
      code_scope.update(_code_scope)

    else:  # constants
      code_scope[f'{target.name}_{split_keys[i]}'] = data
      # replace the data
      if len(split_keys) == i + 1:
        data_to_replace[key] = f'{target.name}_{split_keys[i]}'
      else:
        data_to_replace[key] = f'{target.name}_{split_keys[i]}.{".".join(split_keys[i + 1:])}'

  if has_func_def:
    tree.body[0].decorator_list.clear()
    tree.body[0].args.args.extend([ast.Name(id=a) for a in sorted(arguments)])
    tree.body[0].args.defaults.extend([ast.Constant(None) for _ in sorted(arguments)])
    tree.body[0].args.kwarg = None

  # replace words
  code = tools.ast2code(tree)
  code = tools.word_replace(code, data_to_replace, exclude_dot=True)

  return code, arguments, arg2call, nodes, code_scope


def _replace_this_forloop(tree, iter_name, loop_values, show_code=False, **jit_setting):
  """Replace the given for-loop.

  This function aims to replace the specific for-loop structure, like:

  replace this for-loop

  >>> def update(_t, _dt):
  >>>    for step in self.child_steps.values():
  >>>        step(_t, _dt)

  to

  >>> def update(_t, _dt, AMPA_vec0_delay_g_data=None, AMPA_vec0_delay_g_in_idx=None,
  >>>            AMPA_vec0_delay_g_out_idx=None, AMPA_vec0_s=None, HH0_V=None, HH0_V_th=None,
  >>>            HH0_gNa=None, HH0_h=None, HH0_input=None, HH0_m=None, HH0_n=None, HH0_spike=None):
  >>>    HH0_step(_t, _dt, HH0_V=HH0_V, HH0_V_th=HH0_V_th, HH0_gNa=HH0_gNa,
  >>>             HH0_h=HH0_h, HH0_input=HH0_input, HH0_m=HH0_m, HH0_n=HH0_n,
  >>>             HH0_spike=HH0_spike)
  >>>    AMPA_vec0_step(_t, _dt, AMPA_vec0_delay_g_data=AMPA_vec0_delay_g_data,
  >>>                   AMPA_vec0_delay_g_in_idx=AMPA_vec0_delay_g_in_idx,
  >>>                   AMPA_vec0_delay_g_out_idx=AMPA_vec0_delay_g_out_idx,
  >>>                   AMPA_vec0_s=AMPA_vec0_s, HH0_V=HH0_V, HH0_input=HH0_input,
  >>>                   HH0_spike=HH0_spike)
  >>>    AMPA_vec0_delay_g_step(_t, _dt, AMPA_vec0_delay_g_in_idx=AMPA_vec0_delay_g_in_idx,
  >>>                           AMPA_vec0_delay_g_out_idx=AMPA_vec0_delay_g_out_idx)

  Parameters
  ----------
  tree : ast.Module
    The target code tree.
  iter_name : str
    The for-loop iter.
  loop_values : list/tuple
    The iter contents in the current loop.
  show_code : bool
    Whether show the formatted code.
  """
  assert isinstance(tree, ast.Module)

  replacer = ReplaceThisForLoop(loop_values=loop_values,
                                iter_name=iter_name,
                                show_code=show_code,
                                **jit_setting)
  tree = replacer.visit(tree)
  if not replacer.success:
    raise errors.BrainPyError(f'Do not find the for-loop for "{iter_name}", '
                              f'currently we only support for-loop like '
                              f'"for xxx in {iter_name}:". Does your for-loop '
                              f'structure is not like this. ')

  return tree, replacer.arguments, replacer.arg2call, replacer.nodes, replacer.code_scope


class ReplaceThisForLoop(ast.NodeTransformer):
  def __init__(self, loop_values, iter_name, show_code=False, **jit_setting):
    self.success = False

    # targets
    self.loop_values = loop_values
    self.iter_name = iter_name

    # setting
    self.show_code = show_code
    self.jit_setting = jit_setting

    # results
    self.arguments = set()
    self.arg2call = dict()
    self.nodes = Collector()
    self.code_scope = dict()

  def visit_For(self, node):
    iter_ = tools.ast2code(ast.fix_missing_locations(node.iter))

    if iter_.strip() == self.iter_name:
      data_to_replace = Collector()
      final_node = ast.Module(body=[])
      self.success = True

      # target
      if not isinstance(node.target, ast.Name):
        raise errors.BrainPyError(f'Only support scalar iter, like "for x in xxxx:", not "for '
                                  f'{tools.ast2code(ast.fix_missing_locations(node.target))} '
                                  f'in {iter_}:')
      target = node.target.id

      # for loop values
      for i, value in enumerate(self.loop_values):
        # module and code
        module = ast.Module(body=deepcopy(node).body)
        code = tools.ast2code(module)

        if callable(value):  # transform functions
          r = _jit_func(obj_or_fun=value,
                        show_code=self.show_code,
                        **self.jit_setting)
          tree = _replace_func_call_by_tree(deepcopy(module),
                                            func_call=target,
                                            arg_to_append=r['arguments'],
                                            new_func_name=f'{target}_{i}')

          # update import parameters
          self.arguments.update(r['arguments'])
          self.arg2call.update(r['arg2call'])
          self.nodes.update(r['nodes'])

          # replace the data
          if isinstance(value, Base):
            host = value
            replace_name = f'{host.name}_{target}'
          elif hasattr(value, '__self__') and isinstance(value.__self__, Base):
            host = value.__self__
            replace_name = f'{host.name}_{target}'
          else:
            replace_name = f'{target}_{i}'
          self.code_scope[replace_name] = r['func']
          data_to_replace[f'{target}_{i}'] = replace_name

          final_node.body.extend(tree.body)

        elif isinstance(value, Base):  # transform Base objects
          r = _analyze_cls_func_body(host=value,
                                     self_name=target,
                                     code=code,
                                     tree=module,
                                     show_code=self.show_code,
                                     **self.jit_setting)

          new_code, arguments, arg2call, nodes, code_scope = r
          self.arguments.update(arguments)
          self.arg2call.update(arg2call)
          self.arg2call.update(arg2call)
          self.nodes.update(nodes)
          self.code_scope.update(code_scope)

          final_node.body.extend(ast.parse(new_code).body)

        else:
          raise errors.BrainPyError(f'Only support JIT an iterable objects of function '
                                    f'or Base object, but we got:\n\n {value}')

      # replace words
      final_code = tools.ast2code(final_node)
      final_code = tools.word_replace(final_code, data_to_replace, exclude_dot=True)
      final_node = ast.parse(final_code)

    else:
      final_node = node

    self.generic_visit(final_node)
    return final_node


def _replace_func_call_by_tree(tree, func_call, arg_to_append, remove_self=None,
                               new_func_name=None):
  assert isinstance(func_call, str)
  assert isinstance(arg_to_append, (list, tuple, set))
  assert isinstance(tree, ast.Module)

  transformer = FuncTransformer(func_name=func_call,
                                arg_to_append=arg_to_append,
                                remove_self=remove_self,
                                new_func_name=new_func_name)
  new_tree = transformer.visit(tree)
  return new_tree


def _replace_func_call_by_code(code, func_call, arg_to_append, remove_self=None):
  """Replace functional call.

  This class automatically transform a functional call.
  For example, in your original code:

  >>> V, m, h, n = self.integral(self.V, self.m, self.h, self.n, _t, self.input)

  you want to add new arguments ``gNa``, ``gK`` and ``gL`` into the
  function ``self.integral``. Then this Transformer will help you
  automatically do this:

  >>> V, m, h, n = self.integral(self.V, self.m, self.h, self.n, _t, self.input,
  >>>                            gNa=gNa, gK=gK, gL=gL)

  Parameters
  ----------
  code : str
    The original code string.
  func_call : str
    The functional call.
  arg_to_append : set/list/tuple of str
    The arguments to append.
  remove_self : str, optional
    The self class name to remove.

  Returns
  -------
  new_code : str
    The new code string.
  """
  assert isinstance(func_call, str)
  assert isinstance(arg_to_append, (list, tuple, set))

  tree = ast.parse(code)
  transformer = FuncTransformer(func_name=func_call,
                                arg_to_append=arg_to_append,
                                remove_self=remove_self)
  new_tree = transformer.visit(tree)
  return tools.ast2code(new_tree)


class FuncTransformer(ast.NodeTransformer):
  """Transform a functional call.

  This class automatically transform a functional call.
  For example, in your original code:

  ... code-block:: python

      def update(self, _t, _dt):
        V, m, h, n = self.integral(self.V, self.m, self.h, self.n, _t, self.input)
        self.spike[:] = (self.V < self.V_th) * (V >= self.V_th)

  you want to add new arguments ``gNa``, ``gK`` and ``gL`` into the
  function ``self.integral``. Then this Transformer will help you
  automatically do this:

  ... code-block:: python

      def update(self, _t, _dt):
        V, m, h, n = self.integral(self.V, self.m, self.h, self.n, _t, self.input,
                                   gNa=gNa, gK=gK, gL=gL)
        self.spike[:] = (self.V < self.V_th) * (V >= self.V_th)

  """

  def __init__(self, func_name, arg_to_append, remove_self=None, new_func_name=None):
    self.func_name = func_name
    self.new_func_name = new_func_name
    self.arg_to_append = sorted(arg_to_append)
    self.remove_self = remove_self

  def visit_Call(self, node, level=0):
    if getattr(node, 'starargs', None) is not None:
      raise ValueError("Variable number of arguments (*args) are not supported")
    if getattr(node, 'kwargs', None) is not None:
      raise ValueError("Keyword arguments (**kwargs) are not supported")

    # get function name
    call = tools.ast2code(node.func)
    if call == self.func_name:
      # args
      args = [self.generic_visit(arg) for arg in node.args]
      # remove self arg
      if self.remove_self:
        if args[0].id == self.remove_self:
          args.pop(0)
      # kwargs
      kwargs = [self.generic_visit(keyword) for keyword in node.keywords]
      # new kwargs
      arg_to_append = deepcopy(self.arg_to_append)
      for arg in kwargs:
        if arg.arg in arg_to_append:
          arg_to_append.remove(arg.arg)
      if len(arg_to_append):
        code = f'f({", ".join([f"{k}={k}" for k in arg_to_append])})'
        tree = ast.parse(code)
        new_keywords = tree.body[0].value.keywords
        kwargs.extend(new_keywords)
      # final function
      if self.new_func_name:
        func_call = ast.parse(f'{self.new_func_name}()').body[0].value.func
      else:
        func_call = node.func
      return ast.Call(func=func_call, args=args, keywords=kwargs)
    return node


def _add_try_except(code):
  splits = re.compile(r'\)\s*?:').split(code)
  if len(splits) == 1:
    raise ValueError(f"Cannot analyze code:\n{code}")

  def_line = splits[0] + '):'
  code_lines = '):'.join(splits[1:])
  code_lines = [line for line in code_lines.split('\n') if line.strip()]
  main_code = tools.deindent("\n".join(code_lines))

  code = def_line + '\n'
  code += '  try:\n'
  code += tools.indent(main_code, num_tabs=2, spaces_per_tab=2)
  code += '\n'
  code += '  except NumbaError:\n'
  code += '    print(_code_)'
  return code, {'NumbaError': numba.errors.NumbaError,
                '_code_': code}


def _items2lines(items, num_each_line=1, separator=', ', line_break='\n\t\t'):
  res = ''
  for item in items[:num_each_line]:
    res += item + separator
  for i in range(num_each_line, len(items), num_each_line):
    res += line_break
    for item in items[i: i + num_each_line]:
      res += item + separator
  return res


def _get_args(f):
  # 1. get the function arguments
  original_args = []
  args = []
  kwargs = []

  for name, par in inspect.signature(f).parameters.items():
    if par.kind is inspect.Parameter.POSITIONAL_OR_KEYWORD:
      args.append(par.name)
    elif par.kind is inspect.Parameter.VAR_POSITIONAL:
      args.append(par.name)
    elif par.kind is inspect.Parameter.KEYWORD_ONLY:
      args.append(par.name)
    elif par.kind is inspect.Parameter.POSITIONAL_ONLY:
      raise errors.BrainPyError('Don not support positional only parameters, e.g., /')
    elif par.kind is inspect.Parameter.VAR_KEYWORD:
      kwargs.append(par.name)
    else:
      raise errors.BrainPyError(f'Unknown argument type: {par.kind}')

    original_args.append(str(par))

  # 2. analyze the function arguments
  #   2.1 class keywords
  class_kw = []
  if original_args[0] in profile.CLASS_KEYWORDS:
    class_kw.append(original_args[0])
    original_args = original_args[1:]
    args = args[1:]
  for a in original_args:
    if a.split('=')[0].strip() in profile.CLASS_KEYWORDS:
      raise errors.DiffEqError(f'Class keywords "{a}" must be defined '
                               f'as the first argument.')
  return class_kw, args, kwargs, original_args


def _form_final_call(f_org, f_rep, arg2call, arguments, nodes, show_code=False, name=None):
  _, args, kwargs, org_args = _get_args(f_org)

  name = (name or f_org.__name__)
  code_scope = {key: node for key, node in nodes.items()}
  code_scope[name] = f_rep

  new_args = [] + args
  new_args += [f"{a}={arg2call[a]}" for a in sorted(arguments)]
  new_args += [f"**{a}" for a in kwargs]
  called_args = _items2lines(new_args).strip()
  code_lines = [f'def new_{name}({", ".join(org_args)}):',
                f'  {name}({called_args.strip()})']

  # compile new function
  code = '\n'.join(code_lines)
  # code, _scope = _add_try_except(code)
  # code_scope.update(_scope)
  if show_code:
    show_compiled_codes(code, code_scope)
  exec(compile(code, '', 'exec'), code_scope)
  func = code_scope[f'new_{name}']
  return func


def show_compiled_codes(code, scope):
  print('The recompiled function:')
  print('-------------------------')
  print(code)
  print()
  print('The namespace of the above function:')
  pprint(scope)
  print()


def _find_all_forloop(code_or_tree):
  """Find all for-loops in the code.

  >>> code = '''
  >>> for ch in self._update_channels:
  >>>  ch.update(_t, _dt)
  >>> for ch in self._current_channels:
  >>>  self.input += ch.update(_t, _dt)
  >>> '''
  >>> _find_all_forloop(code)
  {'self._current_channels': ('ch',
                            <_ast.Module object at 0x00000155BD23B730>,
                            'self.input += ch.update(_t, _dt)\n'),
  'self._update_channels': ('ch',
                            <_ast.Module object at 0x00000155B699AD90>,
                            'ch.update(_t, _dt)\n')}

  >>> code = '''
  >>> self.pre_spike.push(self.pre.spike)
  >>> pre_spike = self.pre_spike.pull()
  >>>
  >>> self.g[:] = self.integral(self.g, _t, dt=_dt)
  >>> for pre_id in range(self.pre.num):
  >>>   if pre_spike[pre_id]:
  >>>     start, end = self.pre_slice[pre_id]
  >>>     for post_id in self.post_ids[start: end]:
  >>>       self.g[post_id] += self.g_max
  >>>
  >>> self.post.input[:] += self.output_current(self.g)
  >>>   '''
  >>> _find_all_forloop(code)
  {'range(self.pre.num)': ('pre_id',
                         <_ast.Module object at 0x000001D0AB120D60>,
                         'if pre_spike[pre_id]:\n'
                         '    start, end = self.pre_slice[pre_id]\n'
                         '    for post_id in self.post_ids[start:end]:\n'
                         '        self.g[post_id] += self.g_max\n'),
  'self.post_ids[start:end]': ('post_id',
                              <_ast.Module object at 0x000001D0AB11B460>,
                              'self.g[post_id] += self.g_max\n')}

  Parameters
  ----------
  code_or_tree: str, ast.Module

  Returns
  -------
  res : dict
    with <iter, (target, body, code)>
  """

  # code or tree
  if isinstance(code_or_tree, str):
    code_or_tree = ast.parse(code_or_tree)
  elif isinstance(code_or_tree, ast.Module):
    code_or_tree = code_or_tree
  else:
    raise ValueError

  # finder
  finder = FindAllForLoop()
  finder.visit(code_or_tree)

  # dictionary results
  res = dict()
  for iter_, target, body, body_str in zip(finder.for_iter, finder.for_target,
                                           finder.for_body, finder.for_body_str):
    res[iter_] = (target, body, body_str)
  return res


class FindAllForLoop(ast.NodeVisitor):
  def __init__(self):
    self.for_iter = []
    self.for_target = []
    self.for_body = []
    self.for_body_str = []

  def visit_For(self, node):
    self.for_target.append(tools.ast2code(ast.fix_missing_locations(node.target)))
    self.for_iter.append(tools.ast2code(ast.fix_missing_locations(node.iter)))
    self.for_body.append(ast.Module(body=[deepcopy(r) for r in node.body]))
    codes = tuple(tools.ast2code(ast.fix_missing_locations(r)) for r in node.body)
    self.for_body_str.append('\n'.join(codes))

    self.generic_visit(node)



# -*- coding: utf-8 -*-

from brainpy.math.numpy import ops

import numba.misc.help.inspector as inspector


def test():
  supported = []
  unsupported = []
  for k in dir(ops):
    if not k.startswith('__'):
      v = getattr(ops, k)
      if isinstance(v, type(test)):
        r = inspector.inspect_function(v)
        if r['numba_type'] is None:
          print(v)
          unsupported.append(k)
        else:
          supported.append(k)
  print('Supported:')
  print(supported)
  print('unsupported:')
  print(unsupported)


# -*- coding: utf-8 -*-

from brainpy import math
from brainpy.simulation.brainobjects.base import DynamicalSystem

__all__ = [
  'Dropout'
]


class Dropout(DynamicalSystem):
  """A layer that stochastically ignores a subset of inputs each training step.

  In training, to compensate for the fraction of input values dropped (`rate`),
  all surviving values are multiplied by `1 / (1 - rate)`.

  The parameter `shared_axes` allows to specify a list of axes on which
  the mask will be shared: we will use size 1 on those axes for dropout mask
  and broadcast it. Sharing reduces randomness, but can save memory.

  This layer is active only during training (`mode='train'`). In other
  circumstances it is a no-op.

  Originally introduced in the paper "Dropout: A Simple Way to Prevent Neural
  Networks from Overfitting" available under the following link:
  https://www.cs.toronto.edu/~hinton/absps/JMLRdropout.pdf

  Parameters
  ----------
  prob : float
    Probability to keep element of the tensor.
  steps : tuple of str, tuple of function, dict of (str, function), optional
    The callable function, or a list of callable functions.
  monitors : None, list, tuple, Monitor
    Variables to monitor.
  name : str, optional
    The name of the dynamic system.
  """

  def __init__(self, prob, **kwargs):
    super(Dropout, self).__init__(**kwargs)

    # probability
    self.prob = prob

  def update(self, x, **kwargs):
    if kwargs.get('train', True):
      keep_mask = math.random.bernoulli(self.prob, x.shape)
      return math.where(keep_mask, x / self.prob, 0.)
    else: return x



# -*- coding: utf-8 -*-

import brainpy as bp
from brainpy.integrators.sde.euler_and_milstein import exponential_euler


def test1():
    p = 0.1

    def lorenz_g(x, y, z, t, sigma=10, beta=8 / 3, rho=28):
        return p * x, p * y, p * z

    def lorenz_f(x, y, z, t, sigma=10, beta=8 / 3, rho=28):
        dx = sigma * (y - x)
        dy = x * (rho - z) - y
        dz = x * y - beta * z
        return dx, dy, dz

    exponential_euler(f=lorenz_f, g=lorenz_g, dt=0.01,
                      intg_type=bp.integrators.ITO_SDE,
                      wiener_type=bp.integrators.SCALAR_WIENER,
                      var_type=bp.integrators.POP_VAR,
                      show_code=True)




# -*- coding: utf-8 -*-

from brainpy.simulation.brainobjects.base import Container

__all__ = [
  'Network'
]


class Network(Container):
  """Base class to model network objects, an alias of Container.

  Network instantiates a network, which is aimed to load
  neurons, synapses, and other brain objects.

  Parameters
  ----------
  name : str, Optional
    The network name.
  monitors : optional, list of str, tuple of str
    The items to monitor.
  ds_tuple : 
    A list/tuple container of dynamical system.
  ds_dict : 
    A dict container of dynamical system. 
  """

  def __init__(self, *ds_tuple, monitors=None, name=None, **ds_dict):
    super(Network, self).__init__(*ds_tuple, name=name, monitors=monitors, **ds_dict)



# -*- coding: utf-8 -*-

import logging
import os.path

from brainpy import errors
from brainpy.tools import namechecking
from brainpy.base import collector
from brainpy.base import io

math = DE_INT = None

__all__ = [
  'Base',
]

logger = logging.getLogger('brainpy.base')


class Base(object):
  """The Base class for whole BrainPy ecosystem.

  The subclass of Base includes:

  - ``Module`` in brainpy.dnn.base.py
  - ``DynamicalSystem`` in brainpy.simulation.brainobjects.base.py

  """
  target_backend = None

  def __init__(self, name=None):
    # check whether the object has a unique name.
    self.name = self.unique_name(name=name)
    namechecking.check_name(name=self.name, obj=self)

    # target backend
    if self.target_backend is None:
      self.target_backend = ('general',)
    elif isinstance(self.target_backend, str):
      self.target_backend = (self.target_backend,)
    elif isinstance(self.target_backend, (tuple, list)):
      if not isinstance(self.target_backend[0], str):
        raise errors.BrainPyError('"target_backend" must be a list/tuple of string.')
      self.target_backend = tuple(self.target_backend)
    else:
      raise errors.BrainPyError(f'Unknown setting of "target_backend": {self.target_backend}')

    # check target backend
    global math
    if math is None: from brainpy import math
    check1 = self.target_backend[0] != 'general'
    check2 = math.get_backend_name() not in self.target_backend
    if check1 and check2:
      msg = f'ERROR: The model {self.name} is target to run on {self.target_backend}, ' \
            f'but currently the selected backend is "{math.get_backend_name()}"'
      logger.error(msg)
      raise errors.BrainPyError(msg)

  def vars(self, method='absolute'):
    """Collect all variables in this node and the children nodes.

    Parameters
    ----------
    method : str
      The method to access the variables.

    Returns
    -------
    gather : collector.ArrayCollector
      The collection contained (the path, the variable).
    """
    global math
    if math is None: from brainpy import math

    nodes = self.nodes(method=method)
    gather = collector.ArrayCollector()
    for node_path, node in nodes.items():
      for k in dir(node):
        v = getattr(node, k)
        if isinstance(v, math.Variable):
          gather[f'{node_path}.{k}' if node_path else k] = v
    return gather

  def train_vars(self, method='absolute'):
    """The shortcut for retrieving all trainable variables.

    Parameters
    ----------
    method : str
      The method to access the variables. Support 'absolute' and 'relative'.

    Returns
    -------
    gather : collector.ArrayCollector
      The collection contained (the path, the trainable variable).
    """
    global math
    if math is None:
      from brainpy import math
    return self.vars(method=method).subset(math.TrainVar)

  def nodes(self, method='absolute', _paths=None):
    """Collect all children nodes.

    Parameters
    ----------
    method : str
      The method to access the nodes.
    _paths : set, Optional
      The data structure to solve the circular reference.

    Returns
    -------
    gather : collector.Collector
      The collection contained (the path, the node).
    """
    if _paths is None:
      _paths = set()
    gather = collector.Collector()
    if method == 'absolute':
      nodes = []
      for k, v in self.__dict__.items():
        if isinstance(v, Base):
          path = (id(self), id(v))
          if path not in _paths:
            _paths.add(path)
            gather[v.name] = v
            nodes.append(v)
      for v in nodes:
        gather.update(v.nodes(method=method, _paths=_paths))
      gather[self.name] = self

    elif method == 'relative':
      nodes = []
      gather[''] = self
      for k, v in self.__dict__.items():
        if isinstance(v, Base):
          path = (id(self), id(v))
          if path not in _paths:
            _paths.add(path)
            gather[k] = v
            nodes.append((k, v))
      for k, v in nodes:
        for k2, v2 in v.nodes(method=method, _paths=_paths).items():
          if k2:
            gather[f'{k}.{k2}'] = v2

    else:
      raise ValueError(f'No support for the method of "{method}".')
    return gather

  def _nodes_in_container(self, dict_container, method='absolute', _paths=None):
    if _paths is None:
      _paths = set()

    gather = collector.Collector()
    if method == 'absolute':
      nodes = []
      for _, node in dict_container.items():
        path = (id(self), id(node))
        if path not in _paths:
          _paths.add(path)
          gather[node.name] = node
          nodes.append(node)
      for node in nodes:
        gather[node.name] = node
        gather.update(node.nodes(method=method, _paths=_paths))
      gather[self.name] = self

    elif method == 'relative':
      nodes = []
      gather[''] = self
      for key, node in dict_container.items():
        path = (id(self), id(node))
        if path not in _paths:
          _paths.add(path)
          gather[key] = node
          nodes.append((key, node))
      for key, node in nodes:
        for key2, node2 in node.nodes(method=method, _paths=_paths).items():
          if key2: gather[f'{key}.{key2}'] = node2

    else:
      raise ValueError(f'No support for the method of "{method}".')
    return gather

  def ints(self, method='absolute'):
    """Collect all integrators in this node and the children nodes.

    Parameters
    ----------
    method : str
      The method to access the integrators.

    Returns
    -------
    collector : collector.Collector
      The collection contained (the path, the integrator).
    """
    global DE_INT
    if DE_INT is None:
      from brainpy.integrators.constants import DE_INT

    nodes = self.nodes(method=method)
    gather = collector.Collector()
    for node_path, node in nodes.items():
      for k in dir(node):
        v = getattr(node, k)
        if callable(v) and hasattr(v, '__name__') and v.__name__.startswith(DE_INT):
          gather[f'{node_path}.{k}' if node_path else k] = v
    return gather

  def unique_name(self, name=None, type=None):
    """Get the unique name for this object.

    Parameters
    ----------
    name : str, optional
      The expected name. If None, the default unique name will be returned.
      Otherwise, the provided name will be checked to guarantee its
      uniqueness.
    type : str, optional
      The type of this class, used for object naming.

    Returns
    -------
    name : str
      The unique name for this object.
    """
    if name is None:
      if type is None:
        return namechecking.get_name(type=self.__class__.__name__)
      else:
        return namechecking.get_name(type=type)
    else:
      namechecking.check_name(name=name, obj=self)
      return name

  def load_states(self, filename):
    """Load the model states.

    Parameters
    ----------
    filename : str
      The filename which stores the model states.
    """
    if not os.path.exists(filename):
      raise errors.BrainPyError(f'Cannot find the file path: {filename}')
    if filename.endswith('.hdf5') or filename.endswith('.h5'):
      io.load_h5(filename, target=self)
    if filename.endswith('.pkl'):
      io.load_pkl(filename, target=self)
    if filename.endswith('.npz'):
      io.load_npz(filename, target=self)
    if filename.endswith('.mat'):
      io.load_mat(filename, target=self)
    raise errors.BrainPyError(f'Unknown file format: {filename}. We only supports {io.SUPPORTED_FORMATS}')

  def save_states(self, filename, **setting):
    """Save the model states.

    Parameters
    ----------
    filename : str
      The file name which to store the model states.
    """
    if filename.endswith('.hdf5') or filename.endswith('.h5'):
      io.save_h5(filename, all_vars=self.vars())
    if filename.endswith('.pkl'):
      io.save_pkl(filename, all_vars=self.vars())
    if filename.endswith('.npz'):
      io.save_npz(filename, all_vars=self.vars(), **setting)
    if filename.endswith('.mat'):
      io.save_mat(filename, all_vars=self.vars())
    raise errors.BrainPyError(f'Unknown file format: {filename}. We only supports {io.SUPPORTED_FORMATS}')

  def to(self, devices):
    global math
    if math is None: from brainpy import math

  def cpu(self):
    global math
    if math is None: from brainpy import math

    if math.get_backend_name() == 'jax':
      all_vars = self.vars().unique()
      for data in all_vars.values():
        data[:] = math.asarray(data.value)
        # TODO

  def cuda(self):
    global math
    if math is None: from brainpy import math
    if math.get_backend_name() != 'jax':
      raise errors.BrainPyError(f'Only support to deploy data into "tpu" device in "jax" backend. '
                                f'While currently the selected backend is "{math.get_backend_name()}".')

  def tpu(self):
    global math
    if math is None: from brainpy import math
    if math.get_backend_name() != 'jax':
      raise errors.BrainPyError(f'Only support to deploy data into "tpu" device in "jax" backend. '
                                f'While currently the selected backend is "{math.get_backend_name()}".')



# -*- coding: utf-8 -*-

import pytest
import brainpy as bp


class LIF(bp.NeuGroup):
  pass


def test_check_name():
  LIF(1, name='X_LIF')

  with pytest.raises(bp.errors.UniqueNameError):
    LIF(1, name='X_LIF')








# -*- coding: utf-8 -*-

from brainpy import math
from brainpy.simulation.brainobjects.base import DynamicalSystem
from brainpy.simulation._imports import mjax, jax
from brainpy.simulation.initialize import XavierNormal, Initializer, ZeroInit

__all__ = [
  'Conv2D',
]


def _check_tuple(v):
  if isinstance(v, (tuple, list)):
    return tuple(v)
  elif isinstance(v, int):
    return (v, v)
  else:
    raise ValueError


class Conv2D(DynamicalSystem):
  """Apply a 2D convolution on a 4D-input batch of shape (N,C,H,W).

  Parameters
  ----------
  num_input : int
    The number of channels of the input tensor.
  num_output : int
    The number of channels of the output tensor.
  kernel_size : int, tuple of int
    The size of the convolution kernel, either tuple (height, width)
    or single number if they're the same.
  strides : int, tuple of int
    The convolution strides, either tuple (stride_y, stride_x) or
    single number if they're the same.
  dilations : int, tuple of int
    The spacing between kernel points (also known as astrous convolution),
    either tuple (dilation_y, dilation_x) or single number if they're the same.
  groups : int
    The number of input and output channels group. When groups > 1 convolution
    operation is applied individually for each group. nin and nout must both
    be divisible by groups.
  padding : int, str
    The padding of the input tensor, either "SAME", "VALID" or numerical values
    (low, high).
  w_init : Initializer
    The initializer for convolution kernel (a function that takes in a HWIO
    shape and returns a 4D matrix).
  b_init : Initializer
    The bias initialization.

  steps : tuple of str, tuple of function, dict of (str, function), optional
      The callable function, or a list of callable functions.
  monitors : None, list, tuple, datastructures.Monitor
      Variables to monitor.
  name : str, optional
      The name of the dynamic system.
  """

  def __init__(self, num_input, num_output, kernel_size, strides=1, dilations=1,
               groups=1, padding='SAME', w_init=XavierNormal(), b_init=ZeroInit(),
               has_bias=True, **kwargs):
    super(Conv2D, self).__init__(**kwargs)

    # parameters
    assert num_input % groups == 0, '"nin" should be divisible by groups'
    assert num_output % groups == 0, '"nout" should be divisible by groups'
    self.strides = _check_tuple(strides)
    self.dilations = _check_tuple(dilations)
    if isinstance(padding, str):
      assert padding in ['SAME', 'VALID']
    elif isinstance(padding, tuple):
      assert len(padding) == 2
      for k in padding:
        isinstance(k, int)
    else:
      raise ValueError
    self.padding = padding
    self.groups = groups
    self.has_bias = has_bias

    # weight initialization
    self.b = mjax.TrainVar(b_init((num_output, 1, 1)))
    self.w = mjax.TrainVar(w_init((*_check_tuple(kernel_size), num_input // groups, num_output)))  # HWIO
    self.w_init = w_init
    if has_bias:
      self.b_init = b_init

  def update(self, x, **kwargs):
    nin = self.w.value.shape[2] * self.groups
    assert x.shape[1] == nin, (f'Attempting to convolve an input with {x.shape[1]} input channels '
                               f'when the convolution expects {nin} channels. For reference, '
                               f'self.w.value.shape={self.w.value.shape} and x.shape={x.shape}.')

    y = jax.lax.conv_general_dilated(lhs=x.value if isinstance(x, mjax.JaxArray) else x,
                                     rhs=self.w.value,
                                     window_strides=self.strides,
                                     padding=self.padding,
                                     rhs_dilation=self.dilations,
                                     feature_group_count=self.groups,
                                     dimension_numbers=('NCHW', 'HWIO', 'NCHW'))
    if self.has_bias: y += self.b.value
    return y



# -*- coding: utf-8 -*-

import numpy as np
import numba
import matplotlib.pyplot as plt
from mpl_toolkits.mplot3d import Axes3D
from brainpy.integrators import ode
import brainpy as bp


sigma = 10
beta = 8 / 3
rho = 28


@numba.njit
def lorenz_f(x, y, z, t):
    dx = sigma * (y - x)
    dy = x * (rho - z) - y
    dz = x * y - beta * z
    return dx, dy, dz


def lorenz_system(method, dt=0.01, tol=0.1):
    bp.backend.set('numba')

    integral = numba.njit(method(lorenz_f, show_code=True, tol=tol, adaptive=True))

    times = np.arange(0, 100, 0.01)
    mon1 = []
    mon2 = []
    mon3 = []
    mon4 = []
    x, y, z = 1, 1, 1
    for t in times:
        x, y, z, dt = integral(x, y, z, t, dt)
        mon1.append(x)
        mon2.append(y)
        mon3.append(z)
        mon4.append(dt)
    mon1 = np.array(mon1)
    mon2 = np.array(mon2)
    mon3 = np.array(mon3)
    mon4 = np.array(mon4)

    # fig = plt.figure()
    # ax = fig.gca(projection='3d')
    # plt.plot(mon1, mon2, mon3)
    # ax.set_xlabel('x')
    # ax.set_xlabel('y')
    # ax.set_xlabel('z')
    #
    # fig = plt.figure()
    # plt.plot(mon4)
    #
    # plt.show()


def test():
    lorenz_system(ode.rkf45, dt=0.01, tol=0.001)
    lorenz_system(ode.rkf12, dt=0.01, tol=0.001)
    lorenz_system(ode.rkdp, dt=0.01, tol=0.001)
    lorenz_system(ode.ck, dt=0.01, tol=0.001)
    # lorenz_system(ode.bs, dt=0.01, tol=0.001)
    lorenz_system(ode.heun_euler, dt=0.01, tol=0.001)




# -*- coding: utf-8 -*-

import logging
from copy import deepcopy

import numpy as np

from brainpy import errors, math, tools
from brainpy.analysis import solver
from brainpy.analysis.symbolic import utils

try:
  import sympy
  from brainpy.integrators import analysis_by_sympy
except ModuleNotFoundError:
  sympy = None
  analysis_by_sympy = None

logger = logging.getLogger('brainpy.analysis')

__all__ = [
  'BaseAnalyzer',
  'Base1DAnalyzer',
  'Base2DAnalyzer',
]


class BaseAnalyzer(object):
  """Dynamics Analyzer for Neuron Models.

  This class is a base class which aims for analyze the analysis in
  neuron models. A neuron model is characterized by a series of dynamical
  variables and parameters:

  .. backend::

      {dF \over dt} = F(v_1, v_2, ..., p_1, p_2, ...)

  where :backend:`v_1, v_2` are variables, :backend:`p_1, p_2` are parameters.

  Parameters
  ----------
  model_or_integrals : simulation.Population, function, functions
      A model of the population, the integrator function,
      or a list/tuple of integrator functions.
  target_vars : dict
      The target/dynamical variables.
  fixed_vars : dict
      The fixed variables.
  target_pars : dict, optional
      The parameters which can be dynamical varied.
  pars_update : dict, optional
      The parameters to update.
  numerical_resolution : float, dict
      The resolution for numerical iterative solvers. Default is 0.1. It can set the
      numerical resolution of dynamical variables or dynamical parameters. For example,
      set ``numerical_resolution=0.1`` will generalize it to all variables and parameters;
      set ``numerical_resolution={var1: 0.1, var2: 0.2, par1: 0.1, par2: 0.05}`` will specify
      the particular resolutions to variables and parameters. Moreover, you can also set
      ``numerical_resolution={var1: np.array([...]), var2: 0.1}`` to specify the search points
      need to explore for variable `var1`. This will be useful to set sense search points at some
      inflection points.
  options : dict, optional
      The other setting parameters, which includes:

          perturbation
              float. The small perturbation used to solve the function derivative.
          sympy_solver_timeout
              float, with the unit of second. The maximum  time allowed to use sympy solver
              to get the variable relationship.
          escape_sympy_solver
              bool. Whether escape to use sympy solver, and directly use numerical optimization
              method to solve the nullcline and fixed points.
          lim_scale
              float. The axis limit scale factor. Default is 1.05. The setting means
              the axes will be clipped to ``[var_min * (1-lim_scale)/2, var_max * (var_max-1)/2]``.
  """

  def __init__(self,
               model_or_integrals,
               target_vars,
               fixed_vars=None,
               target_pars=None,
               pars_update=None,
               numerical_resolution=0.1,
               options=None):

    if (sympy is None) or (analysis_by_sympy is None):
      raise errors.PackageMissingError('"SymPy" must be installed for dynamics analysis.')

    # model
    # -----
    if isinstance(model_or_integrals, utils.DynamicModel):
      self.model = model_or_integrals
    elif (isinstance(model_or_integrals, (tuple, list)) and callable(model_or_integrals[0])) or \
        callable(model_or_integrals):
      self.model = utils.transform_integrals_to_model(model_or_integrals)
    else:
      raise ValueError

    # target variables
    # ----------------
    if not isinstance(target_vars, dict):
      raise errors.BrainPyError('"target_vars" must be a dict, with the format of '
                                '{"var1": (var1_min, var1_max)}.')
    self.target_vars = target_vars
    self.dvar_names = list(self.target_vars.keys())
    for key in self.target_vars.keys():
      if key not in self.model.variables:
        raise ValueError(f'{key} is not a dynamical variable in {self.model}.')

    # fixed variables
    # ----------------
    if not isinstance(fixed_vars, dict):
      raise errors.BrainPyError('"fixed_vars" must be a dict with the format '
                                'of {"var1": val1, "var2": val2}.')
    for key in fixed_vars.keys():
      if key not in self.model.variables:
        raise ValueError(f'{key} is not a dynamical variable in {self.model}.')
    self.fixed_vars = fixed_vars

    # check duplicate
    for key in self.fixed_vars.keys():
      if key in self.target_vars:
        raise errors.BrainPyError(f'"{key}" is defined as a target variable in "target_vars", '
                                  f'but also defined as a fixed variable in "fixed_vars".')

    # equations of dynamical variables
    # --------------------------------
    var2eq = {ana.var_name: ana for ana in self.model.analyzers}
    self.target_eqs = tools.DictPlus()
    for key in self.target_vars.keys():
      if key not in var2eq:
        raise errors.BrainPyError(f'target "{key}" is not a dynamical variable.')
      diff_eq = var2eq[key]
      sub_exprs = diff_eq.get_f_expressions(substitute_vars=list(self.target_vars.keys()))
      old_exprs = diff_eq.get_f_expressions(substitute_vars=None)
      self.target_eqs[key] = tools.DictPlus(sub_exprs=sub_exprs,
                                            old_exprs=old_exprs,
                                            diff_eq=diff_eq,
                                            func_name=diff_eq.func_name)

    # parameters to update
    # ---------------------
    if pars_update is None:
      pars_update = dict()
    if not isinstance(pars_update, dict):
      raise errors.BrainPyError('"pars_update" must be a dict with the format '
                                'of {"par1": val1, "par2": val2}.')
    # keys = set(list(pars_update.keys()))
    # keys.update(list(self.model.pars_update.keys()))
    # new_pars_update = {}
    # for key in keys:
    #   if key not in pars_update and key not in fixed_vars:
    #     new_pars_update[key] = self.model.pars_update[key]
    #   else:
    #     new_pars_update[key] = pars_update[key]
    for key in pars_update.keys():
      if (key not in self.model.scopes) and (key not in self.model.parameters):
        raise errors.BrainPyError(f'"{key}" is not a valid parameter in "{self.model}" model.')
    self.pars_update = pars_update

    # dynamical parameters
    # ---------------------
    if target_pars is None:
      target_pars = dict()
    if not isinstance(target_pars, dict):
      raise errors.BrainPyError('"target_pars" must be a dict with the format of {"par1": (val1, val2)}.')
    for key in target_pars.keys():
      if (key not in self.model.scopes) and (key not in self.model.parameters):
        raise errors.BrainPyError(f'"{key}" is not a valid parameter in "{self.model}" model.')
    self.target_pars = target_pars
    self.dpar_names = list(self.target_pars.keys())

    # check duplicate
    for key in self.pars_update.keys():
      if key in self.target_pars:
        raise errors.BrainPyError(f'"{key}" is defined as a target parameter in "target_pars", '
                                  f'but also defined as a fixed parameter in "pars_update".')

    # resolutions for numerical methods
    # ---------------------------------
    self.resolutions = dict()
    if isinstance(numerical_resolution, float):
      for key, lim in self.target_vars.items():
        self.resolutions[key] = np.arange(*lim, numerical_resolution)
      for key, lim in self.target_pars.items():
        self.resolutions[key] = np.arange(*lim, numerical_resolution)
    elif isinstance(numerical_resolution, dict):
      for key in self.dvar_names + self.dpar_names:
        if key not in numerical_resolution:
          raise errors.BrainPyError(f'Must provide the resolution setting of dynamical '
                                    f'variable/parameter "{key}", '
                                    f'but only get {numerical_resolution}.')
        resolution = numerical_resolution[key]
        if isinstance(resolution, float):
          lim = self.target_vars[key] if key in self.target_vars else self.target_pars[key]
          self.resolutions[key] = np.arange(*lim, resolution)
        elif isinstance(resolution, np.ndarray):
          if not np.ndim(resolution) == 1:
            raise errors.BrainPyError(f'resolution must be a 1D vector, but get its '
                                      f'shape with {resolution.shape}.')
          self.resolutions[key] = np.ascontiguousarray(resolution)
        else:
          raise errors.BrainPyError(f'Unknown resolution setting: {key}: {resolution}')
    else:
      raise errors.BrainPyError(f'Unknown resolution type: {type(numerical_resolution)}')

    # a dict to store the analyzed results
    # -------------------------------------
    # 'dxdt' : The differential function ``f`` of the first variable ``x``.
    #          It can be used as ``dxdt(x, y, ...)``.
    # 'dydt' : The differential function ``g`` of the second variable ``y``.
    #          It can be used as ``dydt(x, y, ...)``.
    # 'dfdx' : The derivative of ``f`` by ``x``. It can be used as ``dfdx(x, y, ...)``.
    # 'dfdy' : The derivative of ``f`` by ``y``. It can be used as ``dfdy(x, y, ...)``.
    # 'dgdx' : The derivative of ``g`` by ``x``. It can be used as ``dgdx(x, y, ...)``.
    # 'dgdy' : The derivative of ``g`` by ``y``. It can be used as ``dgdy(x, y, ...)``.
    # 'jacobian' : The jacobian matrix. It can be used as ``jacobian(x, y, ...)``.
    # 'fixed_point' : The fixed point.
    # 'y_by_x_in_y_eq' :
    # 'x_by_y_in_y_eq' :
    # 'y_by_x_in_x_eq' :
    # 'x_by_y_in_x_eq' :
    self.analyzed_results = tools.DictPlus()

    # other settings
    # --------------
    if options is None:
      options = dict()
    self.options = tools.DictPlus()
    self.options['perturbation'] = options.get('perturbation', 1e-6)
    self.options['sympy_solver_timeout'] = options.get('sympy_solver_timeout', 5)  # s
    self.options['escape_sympy_solver'] = options.get('escape_sympy_solver', False)
    self.options['lim_scale'] = options.get('lim_scale', 1.05)


class Base1DAnalyzer(BaseAnalyzer):
  """Neuron analysis analyzer for 1D system.

  It supports the analysis of 1D dynamical system.

  .. backend::

      {dx \over dt} = f(x, t)
  """

  def __init__(self, *args, **kwargs):
    super(Base1DAnalyzer, self).__init__(*args, **kwargs)

    self.x_var = self.dvar_names[0]
    self.x_eq_group = self.target_eqs[self.x_var]

  def get_f_dx(self):
    """Get the derivative function of the first variable. """
    if 'dxdt' not in self.analyzed_results:
      scope = deepcopy(self.pars_update)
      scope.update(self.fixed_vars)
      # scope.update(analysis_by_sympy.get_mapping_scope())
      scope.update(self.x_eq_group.diff_eq.func_scope)
      scope['math'] = math

      argument = ', '.join(self.dvar_names + self.dpar_names)
      func_code = f'def func({argument}):\n'
      for expr in self.x_eq_group.old_exprs[:-1]:
        func_code += f'  {expr.var_name} = {expr.code}\n'
      func_code += f'  return {self.x_eq_group.old_exprs[-1].code}'
      exec(compile(func_code, '', 'exec'), scope)
      func = scope['func']
      self.analyzed_results['dxdt'] = func
    return self.analyzed_results['dxdt']

  def get_f_dfdx(self, origin=True):
    """Get the derivative of ``f`` by variable ``x``. """
    if 'dfdx' not in self.analyzed_results:
      x_var = self.dvar_names[0]
      x_symbol = sympy.Symbol(x_var, real=True)
      x_eq = self.x_eq_group.sub_exprs[-1].code
      x_eq = analysis_by_sympy.str2sympy(x_eq)

      eq_x_scope = deepcopy(self.pars_update)
      eq_x_scope.update(self.fixed_vars)
      # eq_x_scope.update(analysis_by_sympy.get_mapping_scope())
      eq_x_scope.update(self.x_eq_group['diff_eq'].func_scope)
      eq_x_scope['math'] = math

      argument = ', '.join(self.dvar_names + self.dpar_names)
      time_out = self.options.sympy_solver_timeout

      sympy_failed = True
      if not self.options.escape_sympy_solver and not x_eq.contain_unknown_func:
        try:
          logger.info(f'SymPy solve derivative of "{self.x_eq_group.func_name}'
                      f'({argument})" by "{x_var}", ')
          x_eq = x_eq.expr
          f = utils.timeout(time_out)(lambda: sympy.diff(x_eq, x_symbol))
          dfxdx_expr = f()

          # check
          all_vars = set(eq_x_scope.keys())
          all_vars.update(self.dvar_names + self.dpar_names)
          if utils.contain_unknown_symbol(analysis_by_sympy.sympy2str(dfxdx_expr), all_vars):
            logger.info('\tfailed because contain unknown symbols.')
            sympy_failed = True
          else:
            logger.info('\tsuccess.')
            func_codes = [f'def dfdx({argument}):']
            for expr in self.x_eq_group.sub_exprs[:-1]:
              func_codes.append(f'{expr.var_name} = {expr.code}')
            func_codes.append(f'return {analysis_by_sympy.sympy2str(dfxdx_expr)}')
            exec(compile('\n  '.join(func_codes), '', 'exec'), eq_x_scope)
            dfdx = eq_x_scope['dfdx']
            sympy_failed = False
        except KeyboardInterrupt:
          logger.info(f'\tfailed because {time_out} s timeout.')
        except NotImplementedError:
          logger.info('\tfailed because the equation is too complex.')

      if sympy_failed:
        scope = dict(_fx=self.get_f_dx(), perturb=self.options.perturbation, math=math)
        func_codes = [f'def dfdx({argument}):']
        if not origin:
          func_codes.append(f'origin = _fx({argument})')
        func_codes.append(f'disturb = _fx({x_var}+perturb, '
                          f'{",".join(self.dvar_names[1:] + self.dpar_names)})')
        if not origin:
          func_codes.append(f'return (disturb - origin) / perturb')
        else:
          func_codes.append(f'return disturb / perturb')
        exec(compile('\n  '.join(func_codes), '', 'exec'), scope)
        dfdx = scope['dfdx']
      self.analyzed_results['dfdx'] = dfdx
    return self.analyzed_results['dfdx']

  def get_f_fixed_point(self):
    """Get the function to solve the fixed point.
    """

    if 'fixed_point' not in self.analyzed_results:
      x_eq = analysis_by_sympy.str2sympy(self.x_eq_group.sub_exprs[-1].code)

      scope = deepcopy(self.pars_update)
      scope.update(self.fixed_vars)
      # scope.update(analysis_by_sympy.get_mapping_scope())
      scope.update(self.x_eq_group.diff_eq.func_scope)
      scope['math'] = math

      timeout_len = self.options.sympy_solver_timeout
      argument1 = ', '.join(self.dvar_names + self.dpar_names)
      argument2 = ", ".join(self.dvar_names[1:] + self.dpar_names)

      sympy_failed = True
      if not self.options.escape_sympy_solver and not x_eq.contain_unknown_func:
        try:
          logger.info(f'SymPy solve "{self.x_eq_group.func_name}({argument1}) = 0" '
                      f'to "{self.x_var} = f({argument2})", ')

          # solver
          f = utils.timeout(timeout_len)(
            lambda: sympy.solve(x_eq.expr, sympy.Symbol(self.x_var, real=True)))
          results = f()
          for res in results:
            all_vars = set(scope.keys())
            all_vars.update(self.dvar_names + self.dpar_names)
            if utils.contain_unknown_symbol(analysis_by_sympy.sympy2str(res), all_vars):
              logger.info('\tfailed because contain unknown symbols.')
              sympy_failed = True
              break
          else:
            logger.info('\tsuccess.')
            # function codes
            func_codes = [f'def solve_x({argument2}):']
            for expr in self.x_eq_group.sub_exprs[:-1]:
              func_codes.append(f'{expr.var_name} = {expr.code}')
            result_expr = ', '.join([analysis_by_sympy.sympy2str(expr)
                                     for expr in results])
            func_codes.append(f'_res_ = {result_expr}')
            func_codes.append(f'return math.array(_res_)')

            # function compilation
            exec(compile('\n  '.join(func_codes), '', 'exec'), scope)
            self.analyzed_results['fixed_point'] = scope['solve_x']
            sympy_failed = False
        except NotImplementedError:
          logger.info('\tfailed because the equation is too complex.')
          sympy_failed = True
        except KeyboardInterrupt:
          logger.info(f'\tfailed because {timeout_len} s timeout.')
          sympy_failed = True

      if sympy_failed:
        # function codes
        func_codes = [f'def optimizer_x({argument1}):']
        for expr in self.x_eq_group.old_exprs[:-1]:
          func_codes.append(f'{expr.var_name} = {expr.code}')
        func_codes.append(f'return {self.x_eq_group.old_exprs[-1].code}')

        # function compile
        optimizer = utils.jit_compile(scope, '\n  '.join(func_codes), 'optimizer_x')
        xs = self.resolutions[self.x_var]

        def f(*args):
          # `args` corresponds to `self.dvar_names[1:] + self.dpar_names`
          x_values = solver.find_root_of_1d(optimizer, xs, args)
          return np.array(x_values)

        self.analyzed_results['fixed_point'] = f

    return self.analyzed_results['fixed_point']


class Base2DAnalyzer(Base1DAnalyzer):
  """Neuron analysis analyzer for 2D system.

  It supports the analysis of 2D dynamical system.

  .. backend::

      {dx \over dt} = f(x, t, y)

      {dy \over dt} = g(y, t, x)

  Parameters
  ----------

  options : dict, optional
      The other setting parameters, which includes:

          shgo_args
              dict. Arguments of `shgo` optimization method, which can be used to set the
              fields of: constraints, n, iters, callback, minimizer_kwargs, options,
              sampling_method.
          show_shgo
              bool. whether print the shgo's value.
          fl_tol
              float. The tolerance of the function value to recognize it as a condidate of
              function root point.
          xl_tol
              float. The tolerance of the l2 norm distances between this point and previous
              points. If the norm distances are all bigger than `xl_tol` means this
              point belong to a new function root point.

  """

  def __init__(self, *args, **kwargs):
    super(Base2DAnalyzer, self).__init__(*args, **kwargs)

    self.y_var = self.dvar_names[1]
    self.y_eq_group = self.target_eqs[self.y_var]

    # options
    # ---------

    options = kwargs.get('options', dict())
    if options is None:
      options = dict()
    self.options['shgo_args'] = options.get('shgo_args', dict())
    self.options['show_shgo'] = options.get('show_shgo', False)
    self.options['fl_tol'] = options.get('fl_tol', 1e-6)
    self.options['xl_tol'] = options.get('xl_tol', 1e-4)
    for a in ['y_by_x_in_y_eq', 'y_by_x_in_x_eq', 'x_by_y_in_x_eq', 'x_by_y_in_y_eq']:
      if a in options:
        # check "subs"
        subs = options[a]
        if isinstance(subs, str):
          subs = [subs]
        elif isinstance(subs, (tuple, list)):
          subs = subs
        else:
          raise ValueError(f'Unknown setting of "{a}": {subs}')

        # check "f"
        scope = deepcopy(self.pars_update)
        scope.update(self.fixed_vars)
        scope['math'] = math
        # scope.update(analysis_by_sympy.get_mapping_scope())
        if a.endswith('y_eq'):
          scope.update(self.y_eq_group['diff_eq'].func_scope)
        else:
          scope.update(self.x_eq_group['diff_eq'].func_scope)

        # function code
        argument = ",".join(self.dvar_names[2:] + self.dpar_names)
        if a.startswith('y_by_x'):
          func_codes = [f'def func({self.x_var}, {argument}):\n']
        else:
          func_codes = [f'def func({self.y_var}, {argument}):\n']
        func_codes.extend(subs)
        func_codes.append(f'return {subs[-1].split("=")[0]}')

        # function compilation
        exec(compile("\n  ".join(func_codes), '', 'exec'), scope)
        f = scope['func']

        # results
        self.analyzed_results[a] = tools.DictPlus(status='sympy_success', subs=subs, f=f)

  def get_f_dy(self):
    """Get the derivative function of the second variable. """
    if 'dydt' not in self.analyzed_results:
      if len(self.dvar_names) < 2:
        raise errors.BrainPyError(f'Analyzer only receives {len(self.dvar_names)} '
                                  f'dynamical variables, cannot get "dy".')
      y_var = self.dvar_names[1]
      scope = deepcopy(self.pars_update)
      scope.update(self.fixed_vars)
      # scope.update(analysis_by_sympy.get_mapping_scope())
      scope.update(self.y_eq_group.diff_eq.func_scope)
      scope['math'] = math

      argument = ', '.join(self.dvar_names + self.dpar_names)
      func_code = f'def func({argument}):\n'
      for expr in self.y_eq_group.old_exprs[:-1]:
        func_code += f'  {expr.var_name} = {expr.code}\n'
      func_code += f'  return {self.y_eq_group.old_exprs[-1].code}'
      exec(compile(func_code, '', 'exec'), scope)
      self.analyzed_results['dydt'] = scope['func']
    return self.analyzed_results['dydt']

  def get_f_dfdy(self, origin=True):
    """Get the derivative of ``f`` by variable ``y``. """
    if 'dfdy' not in self.analyzed_results:
      x_var = self.dvar_names[0]
      y_var = self.dvar_names[1]
      y_symbol = sympy.Symbol(y_var, real=True)
      x_eq = self.target_eqs[x_var].sub_exprs[-1].code
      x_eq = analysis_by_sympy.str2sympy(x_eq)

      eq_x_scope = deepcopy(self.pars_update)
      eq_x_scope.update(self.fixed_vars)
      # eq_x_scope.update(analysis_by_sympy.get_mapping_scope())
      eq_x_scope.update(self.x_eq_group['diff_eq'].func_scope)
      eq_x_scope['math'] = math

      argument = ', '.join(self.dvar_names + self.dpar_names)
      time_out = self.options.sympy_solver_timeout

      sympy_failed = True
      if not self.options.escape_sympy_solver and not x_eq.contain_unknown_func:
        try:
          logger.info(f'SymPy solve derivative of "{self.x_eq_group.func_name}'
                      f'({argument})" by "{y_var}", ')
          x_eq = x_eq.expr
          f = utils.timeout(time_out)(lambda: sympy.diff(x_eq, y_symbol))
          dfxdy_expr = f()

          # check
          all_vars = set(eq_x_scope.keys())
          all_vars.update(self.dvar_names + self.dpar_names)
          if utils.contain_unknown_symbol(analysis_by_sympy.sympy2str(dfxdy_expr), all_vars):
            logger.info('\tfailed because contain unknown symbols.')
            sympy_failed = True
          else:
            logger.info('\tsuccess.')
            func_codes = [f'def dfdy({argument}):']
            for expr in self.x_eq_group.sub_exprs[:-1]:
              func_codes.append(f'{expr.var_name} = {expr.code}')
            func_codes.append(f'return {analysis_by_sympy.sympy2str(dfxdy_expr)}')
            exec(compile('\n  '.join(func_codes), '', 'exec'), eq_x_scope)
            dfdy = eq_x_scope['dfdy']
            sympy_failed = False
        except KeyboardInterrupt:
          logger.info(f'\tfailed because {time_out} s timeout.')
        except NotImplementedError:
          logger.info('\tfailed because the equation is too complex.')

      if sympy_failed:
        scope = dict(_fx=self.get_f_dx(), perturb=self.options.perturbation, math=math)
        func_codes = [f'def dfdy({argument}):']
        if not origin:
          func_codes.append(f'origin = _fx({argument})')
        func_codes.append(f'disturb = _fx({x_var}, {y_var}+perturb, '
                          f'{",".join(self.dvar_names[2:] + self.dpar_names)})')
        if not origin:
          func_codes.append(f'return (disturb - origin) / perturb')
        else:
          func_codes.append(f'return disturb / perturb')
        exec(compile('\n  '.join(func_codes), '', 'exec'), scope)
        dfdy = scope['dfdy']

      self.analyzed_results['dfdy'] = dfdy
    return self.analyzed_results['dfdy']

  def get_f_dgdx(self, origin=True):
    """Get the derivative of ``g`` by variable ``x``. """
    if 'dgdx' not in self.analyzed_results:
      x_var = self.dvar_names[0]
      x_symbol = sympy.Symbol(x_var, real=True)
      y_var = self.dvar_names[1]
      y_eq = self.target_eqs[y_var].sub_exprs[-1].code
      y_eq = analysis_by_sympy.str2sympy(y_eq)

      eq_y_scope = deepcopy(self.pars_update)
      eq_y_scope.update(self.fixed_vars)
      # eq_y_scope.update(analysis_by_sympy.get_mapping_scope())
      eq_y_scope.update(self.y_eq_group['diff_eq'].func_scope)
      eq_y_scope['math'] = math

      argument = ', '.join(self.dvar_names + self.dpar_names)
      time_out = self.options.sympy_solver_timeout

      sympy_failed = True
      if not self.options.escape_sympy_solver and not y_eq.contain_unknown_func:
        try:
          logger.info(f'SymPy solve derivative of "{self.y_eq_group.func_name}'
                      f'({argument})" by "{x_var}", ')
          y_eq = y_eq.expr
          f = utils.timeout(time_out)(lambda: sympy.diff(y_eq, x_symbol))
          dfydx_expr = f()

          # check
          all_vars = set(eq_y_scope.keys())
          all_vars.update(self.dvar_names + self.dpar_names)
          if utils.contain_unknown_symbol(analysis_by_sympy.sympy2str(dfydx_expr), all_vars):
            logger.info('\tfailed because contain unknown symbols.')
            sympy_failed = True
          else:
            logger.info('\tsuccess.')
            func_codes = [f'def dgdx({argument}):']
            for expr in self.y_eq_group.sub_exprs[:-1]:
              func_codes.append(f'{expr.var_name} = {expr.code}')
            func_codes.append(f'return {analysis_by_sympy.sympy2str(dfydx_expr)}')
            exec(compile('\n  '.join(func_codes), '', 'exec'), eq_y_scope)
            dgdx = eq_y_scope['dgdx']
            sympy_failed = False
        except KeyboardInterrupt:
          logger.info(f'\tfailed because {time_out} s timeout.')
        except NotImplementedError:
          logger.info('\tfailed because the equation is too complex.')

      if sympy_failed:
        scope = dict(_fy=self.get_f_dy(), perturb=self.options.perturbation, math=math)
        func_codes = [f'def dgdx({argument}):']
        if not origin:
          func_codes.append(f'origin = _fy({argument})')
        func_codes.append(f'disturb = _fy({x_var}+perturb, '
                          f'{",".join(self.dvar_names[1:] + self.dpar_names)})')
        if not origin:
          func_codes.append(f'return (disturb - origin) / perturb')
        else:
          func_codes.append(f'return disturb / perturb')
        exec(compile('\n  '.join(func_codes), '', 'exec'), scope)
        dgdx = scope['dgdx']

      self.analyzed_results['dgdx'] = dgdx
    return self.analyzed_results['dgdx']

  def get_f_dgdy(self, origin=True):
    """Get the derivative of ``g`` by variable ``y``. """
    if 'dgdy' not in self.analyzed_results:
      x_var = self.dvar_names[0]
      y_var = self.dvar_names[1]
      y_symbol = sympy.Symbol(y_var, real=True)
      y_eq = self.target_eqs[y_var].sub_exprs[-1].code
      y_eq = analysis_by_sympy.str2sympy(y_eq)

      eq_y_scope = deepcopy(self.pars_update)
      eq_y_scope.update(self.fixed_vars)
      # eq_y_scope.update(analysis_by_sympy.get_mapping_scope())
      eq_y_scope.update(self.y_eq_group['diff_eq'].func_scope)
      eq_y_scope['math'] = math

      argument = ', '.join(self.dvar_names + self.dpar_names)
      argument2 = ', '.join(self.dvar_names[2:] + self.dpar_names)
      time_out = self.options.sympy_solver_timeout

      sympy_failed = True
      if not self.options.escape_sympy_solver and not y_eq.contain_unknown_func:
        try:
          logger.info(f'\tSymPy solve derivative of "{self.y_eq_group.func_name}'
                      f'({argument})" by "{y_var}", ')
          y_eq = y_eq.expr
          f = utils.timeout(time_out)(lambda: sympy.diff(y_eq, y_symbol))
          dfydx_expr = f()

          # check
          all_vars = set(eq_y_scope.keys())
          all_vars.update(self.dvar_names + self.dpar_names)
          if utils.contain_unknown_symbol(analysis_by_sympy.sympy2str(dfydx_expr), all_vars):
            logger.info('\tfailed because contain unknown symbols.')
            sympy_failed = True
          else:
            logger.info('\tsuccess.')
            func_codes = [f'def dgdy({argument}):']
            for expr in self.y_eq_group.sub_exprs[:-1]:
              func_codes.append(f'{expr.var_name} = {expr.code}')
            func_codes.append(f'return {analysis_by_sympy.sympy2str(dfydx_expr)}')
            exec(compile('\n  '.join(func_codes), '', 'exec'), eq_y_scope)
            dgdy = eq_y_scope['dgdy']
            sympy_failed = False
        except KeyboardInterrupt:
          logger.info(f'\tfailed because {time_out} s timeout.')
        except NotImplementedError:
          logger.info('\tfailed because the equation is too complex.')

      if sympy_failed:
        scope = dict(_fy=self.get_f_dy(), perturb=self.options.perturbation, math=math)
        func_codes = [f'def dgdy({argument}):']
        if not origin:
          func_codes.append(f'origin = _fy({argument})')
        func_codes.append(f'disturb = _fy({x_var}, {y_var}+perturb, {argument2})')
        if not origin:
          func_codes.append(f'return (disturb - origin) / perturb')
        else:
          func_codes.append(f'return disturb / perturb')
        exec(compile('\n  '.join(func_codes), '', 'exec'), scope)
        dgdy = scope['dgdy']

      self.analyzed_results['dgdy'] = dgdy
    return self.analyzed_results['dgdy']

  def get_f_jacobian(self):
    """Get the function to solve jacobian matrix.
    """
    if 'jacobian' not in self.analyzed_results:
      dfdx = self.get_f_dfdx()
      dfdy = self.get_f_dfdy()
      dgdx = self.get_f_dgdx()
      dgdy = self.get_f_dgdy()

      argument = ','.join(self.dvar_names + self.dpar_names)
      scope = dict(f_dfydy=dgdy, f_dfydx=dgdx, f_dfxdy=dfdy, f_dfxdx=dfdx, math=math)
      func_codes = [f'def f_jacobian({argument}):']
      func_codes.append(f'dfxdx = f_dfxdx({argument})')
      func_codes.append(f'dfxdy = f_dfxdy({argument})')
      func_codes.append(f'dfydx = f_dfydx({argument})')
      func_codes.append(f'dfydy = f_dfydy({argument})')
      func_codes.append('return math.array([[dfxdx, dfxdy], [dfydx, dfydy]])')
      exec(compile('\n  '.join(func_codes), '', 'exec'), scope)
      self.analyzed_results['jacobian'] = scope['f_jacobian']

    return self.analyzed_results['jacobian']

  def get_f_fixed_point(self):
    """Get the function to solve the fixed point.
    """

    if 'fixed_point' not in self.analyzed_results:
      vars_and_pars = ','.join(self.dvar_names[2:] + self.dpar_names)

      eq_xy_scope = deepcopy(self.pars_update)
      eq_xy_scope.update(self.fixed_vars)
      # eq_xy_scope.update(analysis_by_sympy.get_mapping_scope())
      eq_xy_scope.update(self.x_eq_group['diff_eq'].func_scope)
      eq_xy_scope.update(self.y_eq_group['diff_eq'].func_scope)
      eq_xy_scope['math'] = math

      # Try 1: substitute y_group to x_group
      #        y_by_x
      # ------------------------------------
      y_by_x_in_y_eq = self.get_y_by_x_in_y_eq()
      if y_by_x_in_y_eq['status'] == 'sympy_success':
        func_codes = [f'def optimizer_x({self.x_var}, {vars_and_pars}):']
        func_codes += y_by_x_in_y_eq['subs']
        func_codes.extend([f'{expr.var_name} = {expr.code}'
                           for expr in self.x_eq_group.old_exprs[:-1]])
        func_codes.append(f'return {self.x_eq_group.old_exprs[-1].code}')
        func_code = '\n  '.join(func_codes)
        optimizer = utils.jit_compile(eq_xy_scope, func_code, 'optimizer_x')

        def f(*args):
          # ``args`` are equal to ``vars_and_pars``
          x_range = self.resolutions[self.x_var]
          x_values = solver.find_root_of_1d(optimizer, x_range, args)
          x_values = np.array(x_values)
          y_values = y_by_x_in_y_eq['f'](x_values, *args)
          y_values = np.array(y_values)
          return x_values, y_values

        self.analyzed_results['fixed_point'] = f
        return f

      # Try 2: substitute y_group to x_group
      #        x_by_y
      # ------------------------------------
      x_by_y_in_y_eq = self.get_x_by_y_in_y_eq()
      if x_by_y_in_y_eq['status'] == 'sympy_success':
        func_codes = [f'def optimizer_y({self.y_var}, {vars_and_pars}):']
        func_codes += x_by_y_in_y_eq['subs']
        func_codes.extend([f'{expr.var_name} = {expr.code}'
                           for expr in self.x_eq_group.old_exprs[:-1]])
        func_codes.append(f'return {self.x_eq_group.old_exprs[-1].code}')
        func_code = '\n  '.join(func_codes)
        optimizer = utils.jit_compile(eq_xy_scope, func_code, 'optimizer_y')

        def f(*args):
          # ``args`` are equal to ``vars_and_pars``
          y_range = self.resolutions[self.y_var]
          y_values = solver.find_root_of_1d(optimizer, y_range, args)
          y_values = np.array(y_values)
          x_values = x_by_y_in_y_eq['f'](y_values, *args)
          x_values = np.array(x_values)
          return x_values, y_values

        self.analyzed_results['fixed_point'] = f
        return f

      # Try 3: substitute x_group to y_group
      #        y_by_x
      # ------------------------------------
      y_by_x_in_x_eq = self.get_y_by_x_in_x_eq()
      if y_by_x_in_x_eq['status'] == 'sympy_success':
        func_codes = [f'def optimizer_x({self.x_var}, {vars_and_pars}):']
        func_codes += y_by_x_in_x_eq['subs']
        func_codes.extend([f'{expr.var_name} = {expr.code}'
                           for expr in self.y_eq_group.old_exprs[:-1]])
        func_codes.append(f'return {self.y_eq_group.old_exprs[-1].code}')
        func_code = '\n  '.join(func_codes)
        optimizer = utils.jit_compile(eq_xy_scope, func_code, 'optimizer_x')

        def f(*args):
          # ``args`` are equal to ``vars_and_pars``
          x_range = self.resolutions[self.x_var]
          x_values = solver.find_root_of_1d(optimizer, x_range, args)
          x_values = np.array(x_values)
          y_values = y_by_x_in_x_eq['f'](x_values, *args)
          y_values = np.array(y_values)
          return x_values, y_values

        self.analyzed_results['fixed_point'] = f
        return f

      # Try 4: substitute x_group to y_group
      #        x_by_y
      # ------------------------------------
      x_by_y_in_x_eq = self.get_x_by_y_in_x_eq()
      if x_by_y_in_x_eq['status'] == 'sympy_success':
        func_codes = [f'def optimizer_y({self.y_var}, {vars_and_pars}):']
        func_codes += x_by_y_in_x_eq['subs']
        func_codes.extend([f'{expr.var_name} = {expr.code}'
                           for expr in self.y_eq_group.old_exprs[:-1]])
        func_codes.append(f'return {self.y_eq_group.old_exprs[-1].code}')
        func_code = '\n  '.join(func_codes)
        optimizer = utils.jit_compile(eq_xy_scope, func_code, 'optimizer_y')

        def f(*args):
          # ``args`` are equal to ``vars_and_pars``
          y_range = self.resolutions[self.y_var]
          y_values = solver.find_root_of_1d(optimizer, y_range, args)
          y_values = np.array(y_values)
          x_values = x_by_y_in_x_eq['f'](y_values, *args)
          x_values = np.array(x_values)
          return x_values, y_values

        self.analyzed_results['fixed_point'] = f
        return f

      # Try 5: numerical optimization method
      # ------------------------------------
      # f
      eq_x_scope = deepcopy(self.pars_update)
      eq_x_scope.update(self.fixed_vars)
      # eq_x_scope.update(analysis_by_sympy.get_mapping_scope())
      eq_x_scope.update(self.x_eq_group['diff_eq'].func_scope)
      eq_x_scope['math'] = math

      func_codes = [f'def f_x({",".join(self.dvar_names + self.dpar_names)}):']
      func_codes.extend([f'{expr.var_name} = {expr.code}'
                         for expr in self.x_eq_group.old_exprs[:-1]])
      func_codes.append(f'return {self.x_eq_group.old_exprs[-1].code}')
      exec(compile('\n  '.join(func_codes), '', 'exec'), eq_x_scope)
      f_x = eq_x_scope['f_x']

      # g
      eq_y_scope = deepcopy(self.pars_update)
      eq_y_scope.update(self.fixed_vars)
      # eq_y_scope.update(analysis_by_sympy.get_mapping_scope())
      eq_y_scope.update(self.y_eq_group['diff_eq'].func_scope)
      eq_y_scope['math'] = math

      func_codes = [f'def g_y({",".join(self.dvar_names + self.dpar_names)}):']
      func_codes.extend([f'{expr.var_name} = {expr.code}'
                         for expr in self.y_eq_group.old_exprs[:-1]])
      func_codes.append(f'return {self.y_eq_group.old_exprs[-1].code}')
      exec(compile('\n  '.join(func_codes), '', 'exec'), eq_y_scope)
      g_y = eq_y_scope['g_y']

      # f**2 + g**2
      def optimizer(x, *args):
        return f_x(x[0], x[1], *args) ** 2 + g_y(x[0], x[1], *args) ** 2

      # optimization results
      def f(*args):
        # ``args`` are equal to ``vars_and_pars``
        return solver.find_root_of_2d(optimizer,
                                      x_bound=self.target_vars[self.x_var],
                                      y_bound=self.target_vars[self.y_var],
                                      args=args,
                                      shgo_args=self.options.shgo_args,
                                      fl_tol=self.options.fl_tol,
                                      xl_tol=self.options.xl_tol,
                                      verbose=self.options.show_shgo)

      self.analyzed_results['fixed_point'] = f
      return f

    return self.analyzed_results['fixed_point']

  def get_f_optimize_x_nullcline(self, coords=None):
    """Get the function to solve X nullcline by using numerical optimization method.

    Parameters
    ----------
    coords : str
        The coordination.
    """
    if coords is None:
      coords = self.x_var + '-' + self.y_var

    key = f'optimize_x_nullcline,{coords}'
    if key not in self.analyzed_results:
      # check coordinate setting
      coord_splits = [a.strip() for a in coords.strip().split('-')]
      if self.x_var not in coord_splits:
        raise ValueError(f'Variable "{self.x_var}" must be in coordinate '
                         f'settings. But we get "{coords}".')
      if self.y_var not in coord_splits:
        raise ValueError(f'Variable "{self.y_var}" must be in coordinate '
                         f'settings. But we get "{coords}".')

      # x equation scope
      eq_x_scope = deepcopy(self.pars_update)
      eq_x_scope.update(self.fixed_vars)
      # eq_x_scope.update(analysis_by_sympy.get_mapping_scope())
      eq_x_scope.update(self.x_eq_group.diff_eq.func_scope)
      eq_x_scope['math'] = math

      argument = ','.join(self.dvar_names[2:] + self.dpar_names)

      # optimization function
      func_codes = [f'def optimizer_x({self.x_var},{self.y_var},{argument}):']
      for expr in self.x_eq_group.old_exprs[:-1]:
        func_codes.append(f'{expr.var_name} = {expr.code}')
      func_codes.append(f'return {self.x_eq_group.old_exprs[-1].code}')
      func_code = '\n  '.join(func_codes)
      optimizer_x_by_y = utils.jit_compile(eq_x_scope, func_code, 'optimizer_x')

      func_codes = [f'def optimizer_y({self.y_var},{self.x_var},{argument}):']
      for expr in self.x_eq_group.old_exprs[:-1]:
        func_codes.append(f'{expr.var_name} = {expr.code}')
      func_codes.append(f'return {self.x_eq_group.old_exprs[-1].code}')
      func_code = '\n  '.join(func_codes)
      optimizer_y_by_x = utils.jit_compile(eq_x_scope, func_code, 'optimizer_y')

      # optimization results
      xs = self.resolutions[self.x_var]
      ys = self.resolutions[self.y_var]

      def f1(*args):
        # ``args`` corresponds to the dynamical parameters
        x_values, y_values = [], []
        for y in ys:
          for x in solver.find_root_of_1d(optimizer_x_by_y, xs, (y,) + args):
            x_values.append(x)
            y_values.append(y)
        return np.array(x_values), np.array(y_values)

      def f2(*args):
        # ``args`` corresponds to the dynamical parameters
        x_values, y_values = [], []
        for x in xs:
          for y in solver.find_root_of_1d(optimizer_y_by_x, ys, (x,) + args):
            x_values.append(x)
            y_values.append(y)
        return np.array(x_values), np.array(y_values)

      key1 = f'optimize_x_nullcline,{self.x_var}-{self.y_var}'
      key2 = f'optimize_x_nullcline,{self.y_var}-{self.x_var}'
      self.analyzed_results[key1] = f1
      self.analyzed_results[key2] = f2

    return self.analyzed_results[key]

  def get_f_optimize_y_nullcline(self, coords=None):
    """Get the function to solve Y nullcline by using numerical optimization method.

    Parameters
    ----------
    coords : str
        The coordination.
    """
    if coords is None:
      coords = self.x_var + '-' + self.y_var

    key = f'optimize_y_nullcline,{coords}'
    if key not in self.analyzed_results:
      # check coordinate setting
      coord_splits = [a.strip() for a in coords.strip().split('-')]
      if self.x_var not in coord_splits:
        raise ValueError(f'Variable "{self.x_var}" must be in coordinate '
                         f'settings. But we get "{coords}".')
      if self.y_var not in coord_splits:
        raise ValueError(f'Variable "{self.y_var}" must be in coordinate '
                         f'settings. But we get "{coords}".')

      # y equation scope
      eq_y_scope = deepcopy(self.pars_update)
      eq_y_scope.update(self.fixed_vars)
      # eq_y_scope.update(analysis_by_sympy.get_mapping_scope())
      eq_y_scope.update(self.y_eq_group.diff_eq.func_scope)
      eq_y_scope['math'] = math

      argument = ','.join(self.dvar_names[2:] + self.dpar_names)

      # optimization function
      func_codes = [f'def optimizer_x({self.x_var},{self.y_var},{argument}):']
      for expr in self.y_eq_group.old_exprs[:-1]:
        func_codes.append(f'{expr.var_name} = {expr.code}')
      func_codes.append(f'return {self.y_eq_group.old_exprs[-1].code}')
      func_code = '\n  '.join(func_codes)
      optimizer_x_by_y = utils.jit_compile(eq_y_scope, func_code, 'optimizer_x')

      func_codes = [f'def optimizer_y({self.y_var},{self.x_var},{argument}):']
      for expr in self.y_eq_group.old_exprs[:-1]:
        func_codes.append(f'{expr.var_name} = {expr.code}')
      func_codes.append(f'return {self.y_eq_group.old_exprs[-1].code}')
      func_code = '\n  '.join(func_codes)
      optimizer_y_by_x = utils.jit_compile(eq_y_scope, func_code, 'optimizer_y')

      # optimization results
      xs = self.resolutions[self.x_var]
      ys = self.resolutions[self.y_var]

      def f1(*args):
        # ``args`` corresponds to the dynamical parameters
        x_values, y_values = [], []
        for y in ys:
          for x in solver.find_root_of_1d(optimizer_x_by_y, xs, (y,) + args):
            x_values.append(x)
            y_values.append(y)
        return np.array(x_values), np.array(y_values)

      def f2(*args):
        # ``args`` corresponds to the dynamical parameters
        x_values, y_values = [], []
        for x in xs:
          for y in solver.find_root_of_1d(optimizer_y_by_x, ys, (x,) + args):
            x_values.append(x)
            y_values.append(y)
        return np.array(x_values), np.array(y_values)

      key1 = f'optimize_y_nullcline,{self.x_var}-{self.y_var}'
      key2 = f'optimize_y_nullcline,{self.y_var}-{self.x_var}'
      self.analyzed_results[key1] = f1
      self.analyzed_results[key2] = f2

    return self.analyzed_results[key]

  def get_y_by_x_in_y_eq(self):
    """Get the expression of "y_by_x_in_y_eq".

    Specifically, ``self.analyzed_results['y_by_x_in_y_eq']`` is a Dict,
    with the following keywords:

    - status : 'sympy_success', 'sympy_failed', 'escape'
    - subs : substituted expressions (relationship) of y_by_x
    - f : function of y_by_x
    """
    if 'y_by_x_in_y_eq' not in self.analyzed_results:
      results = tools.DictPlus()
      if not self.options.escape_sympy_solver:
        y_symbol = sympy.Symbol(self.y_var, real=True)
        code = self.target_eqs[self.y_var].sub_exprs[-1].code
        y_eq = analysis_by_sympy.str2sympy(code).expr

        eq_y_scope = deepcopy(self.pars_update)
        eq_y_scope.update(self.fixed_vars)
        # eq_y_scope.update(analysis_by_sympy.get_mapping_scope())
        eq_y_scope.update(self.y_eq_group['diff_eq'].func_scope)
        eq_y_scope['math'] = math

        argument = ', '.join(self.dvar_names + self.dpar_names)
        timeout_len = self.options.sympy_solver_timeout

        try:
          logger.info(f'SymPy solve "{self.y_eq_group.func_name}({argument}) = 0" to '
                      f'"{self.y_var} = f({self.x_var}, '
                      f'{",".join(self.dvar_names[2:] + self.dpar_names)})", ')
          # solve the expression
          f = utils.timeout(timeout_len)(lambda: sympy.solve(y_eq, y_symbol))
          y_by_x_in_y_eq = f()
          if len(y_by_x_in_y_eq) > 1:
            raise NotImplementedError('Do not support multiple values.')
          y_by_x_in_y_eq = analysis_by_sympy.sympy2str(y_by_x_in_y_eq[0])

          # check
          all_vars = set(eq_y_scope.keys())
          all_vars.update(self.dvar_names + self.dpar_names)
          if utils.contain_unknown_symbol(y_by_x_in_y_eq, all_vars):
            logger.info('\tfailed because contain unknown symbols.')
            results['status'] = 'sympy_failed'
          else:
            logger.info('\tsuccess.')
            # substituted codes
            subs_codes = [f'{expr.var_name} = {expr.code}'
                          for expr in self.y_eq_group.sub_exprs[:-1]]
            subs_codes.append(f'{self.y_var} = {y_by_x_in_y_eq}')

            # compile the function
            func_code = f'def func({self.x_var}, {",".join(self.dvar_names[2:] + self.dpar_names)}):\n'
            for expr in self.y_eq_group.sub_exprs[:-1]:
              func_code += f'  {expr.var_name} = {expr.code}\n'
            func_code += f'  return {y_by_x_in_y_eq}'
            exec(compile(func_code, '', 'exec'), eq_y_scope)

            # set results
            results['status'] = 'sympy_success'
            results['subs'] = subs_codes
            results['f'] = eq_y_scope['func']

        except NotImplementedError:
          logger.info('\tfailed because the equation is too complex.')
          results['status'] = 'sympy_failed'
        except KeyboardInterrupt:
          logger.info(f'\tfailed because {timeout_len} s timeout.')
          results['status'] = 'sympy_failed'
      else:
        results['status'] = 'escape'
      self.analyzed_results['y_by_x_in_y_eq'] = results
    return self.analyzed_results['y_by_x_in_y_eq']

  def get_y_by_x_in_x_eq(self):
    """Get the expression of "y_by_x_in_x_eq".

    Specifically, ``self.analyzed_results['y_by_x_in_x_eq']`` is a Dict,
    with the following keywords:

    - status : 'sympy_success', 'sympy_failed', 'escape'
    - subs : substituted expressions (relationship) of y_by_x
    - f : function of y_by_x
    """
    if 'y_by_x_in_x_eq' not in self.analyzed_results:
      results = tools.DictPlus()

      if not self.options.escape_sympy_solver:
        y_symbol = sympy.Symbol(self.y_var, real=True)
        code = self.x_eq_group.sub_exprs[-1].code
        x_eq = analysis_by_sympy.str2sympy(code).expr

        eq_x_scope = deepcopy(self.pars_update)
        eq_x_scope.update(self.fixed_vars)
        # eq_x_scope.update(analysis_by_sympy.get_mapping_scope())
        eq_x_scope.update(self.x_eq_group['diff_eq'].func_scope)
        eq_x_scope['math'] = math

        argument = ', '.join(self.dvar_names + self.dpar_names)
        timeout_len = self.options.sympy_solver_timeout

        try:
          logger.info(f'SymPy solve "{self.x_eq_group.func_name}({argument}) = 0" to '
                      f'"{self.y_var} = f({self.x_var}, '
                      f'{",".join(self.dvar_names[2:] + self.dpar_names)})", ')

          # solve the expression
          f = utils.timeout(timeout_len)(lambda: sympy.solve(x_eq, y_symbol))
          y_by_x_in_x_eq = f()
          if len(y_by_x_in_x_eq) > 1:
            raise NotImplementedError('Do not support multiple values.')
          y_by_x_in_x_eq = analysis_by_sympy.sympy2str(y_by_x_in_x_eq[0])

          all_vars = set(eq_x_scope.keys())
          all_vars.update(self.dvar_names + self.dpar_names)
          if utils.contain_unknown_symbol(y_by_x_in_x_eq, all_vars):
            logger.info('\tfailed because contain unknown symbols.')
            results['status'] = 'sympy_failed'
          else:
            logger.info('\tsuccess.')

            # substituted codes
            subs_codes = [f'{expr.var_name} = {expr.code}'
                          for expr in self.x_eq_group.sub_exprs[:-1]]
            subs_codes.append(f'{self.y_var} = {y_by_x_in_x_eq}')

            # compile the function
            func_code = f'def func({self.x_var}, {",".join(self.dvar_names[2:] + self.dpar_names)}):\n'
            for expr in self.y_eq_group.sub_exprs[:-1]:
              func_code += f'  {expr.var_name} = {expr.code}\n'
            func_code += f'  return {y_by_x_in_x_eq}'
            exec(compile(func_code, '', 'exec'), eq_x_scope)

            # set results
            results['status'] = 'sympy_success'
            results['subs'] = subs_codes
            results['f'] = eq_x_scope['func']
        except NotImplementedError:
          logger.info('\tfailed because the equation is too complex.')
          results['status'] = 'sympy_failed'
        except KeyboardInterrupt:
          logger.info(f'\tfailed because {timeout_len} s timeout.')
          results['status'] = 'sympy_failed'
      else:
        results['status'] = 'escape'
      self.analyzed_results['y_by_x_in_x_eq'] = results
    return self.analyzed_results['y_by_x_in_x_eq']

  def get_x_by_y_in_y_eq(self):
    """Get the expression of "x_by_y_in_y_eq".

    Specifically, ``self.analyzed_results['x_by_y_in_y_eq']`` is a Dict,
    with the following keywords:

    - status : 'sympy_success', 'sympy_failed', 'escape'
    - subs : substituted expressions (relationship) of x_by_y
    - f : function of x_by_y
    """
    if 'x_by_y_in_y_eq' not in self.analyzed_results:
      results = tools.DictPlus()

      if not self.options.escape_sympy_solver:
        x_symbol = sympy.Symbol(self.x_var, real=True)
        code = self.target_eqs[self.y_var].sub_exprs[-1].code
        y_eq = analysis_by_sympy.str2sympy(code).expr

        eq_y_scope = deepcopy(self.pars_update)
        eq_y_scope.update(self.fixed_vars)
        # eq_y_scope.update(analysis_by_sympy.get_mapping_scope())
        eq_y_scope.update(self.y_eq_group['diff_eq'].func_scope)
        eq_y_scope['math'] = math

        argument = ', '.join(self.dvar_names + self.dpar_names)
        timeout_len = self.options.sympy_solver_timeout

        try:
          logger.info(f'SymPy solve "{self.y_eq_group.func_name}({argument}) = 0" to '
                      f'"{self.x_var} = f({",".join(self.dvar_names[1:] + self.dpar_names)})", ')
          # solve the expression
          f = utils.timeout(timeout_len)(lambda: sympy.solve(y_eq, x_symbol))
          x_by_y_in_y_eq = f()
          if len(x_by_y_in_y_eq) > 1:
            raise NotImplementedError('Do not support multiple values.')
          x_by_y_in_y_eq = analysis_by_sympy.sympy2str(x_by_y_in_y_eq[0])

          # check
          all_vars = set(eq_y_scope.keys())
          all_vars.update(self.dvar_names + self.dpar_names)
          if utils.contain_unknown_symbol(x_by_y_in_y_eq, all_vars):
            logger.info('\tfailed because contain unknown symbols.')
            results['status'] = 'sympy_failed'
          else:
            logger.info('\tsuccess.')

            # substituted codes
            subs_codes = [f'{expr.var_name} = {expr.code}'
                          for expr in self.y_eq_group.sub_exprs[:-1]]
            subs_codes.append(f'{self.x_var} = {x_by_y_in_y_eq}')

            # compile the function
            func_code = f'def func({",".join(self.dvar_names[1:] + self.dpar_names)}):\n'
            for expr in self.y_eq_group.sub_exprs[:-1]:
              func_code += f'  {expr.var_name} = {expr.code}\n'
            func_code += f'  return {x_by_y_in_y_eq}'
            exec(compile(func_code, '', 'exec'), eq_y_scope)

            # set results
            results['status'] = 'sympy_success'
            results['subs'] = subs_codes
            results['f'] = eq_y_scope['func']
        except NotImplementedError:
          logger.info('\tfailed because the equation is too complex.')
          results['status'] = 'sympy_failed'
        except KeyboardInterrupt:
          logger.info(f'\tfailed because {timeout_len} s timeout.')
          results['status'] = 'sympy_failed'
      else:
        results['status'] = 'escape'
      self.analyzed_results['x_by_y_in_y_eq'] = results
    return self.analyzed_results['x_by_y_in_y_eq']

  def get_x_by_y_in_x_eq(self):
    """Get the expression of "x_by_y_in_x_eq".

    Specifically, ``self.analyzed_results['x_by_y_in_x_eq']`` is a Dict,
    with the following keywords:

    - status : 'sympy_success', 'sympy_failed', 'escape'
    - subs : substituted expressions (relationship) of x_by_y
    - f : function of x_by_y
    """
    if 'x_by_y_in_x_eq' not in self.analyzed_results:
      results = tools.DictPlus()

      if not self.options.escape_sympy_solver:
        x_symbol = sympy.Symbol(self.x_var, real=True)
        code = self.x_eq_group.sub_exprs[-1].code
        x_eq = analysis_by_sympy.str2sympy(code).expr

        eq_x_scope = deepcopy(self.pars_update)
        eq_x_scope.update(self.fixed_vars)
        # eq_x_scope.update(analysis_by_sympy.get_mapping_scope())
        eq_x_scope.update(self.x_eq_group['diff_eq'].func_scope)
        eq_x_scope['math'] = math

        argument = ', '.join(self.dvar_names + self.dpar_names)
        timeout_len = self.options.sympy_solver_timeout

        try:
          logger.info(f'SymPy solve "{self.x_eq_group.func_name}({argument}) = 0" to '
                      f'"{self.x_var} = f({",".join(self.dvar_names[1:] + self.dpar_names)})", ')
          # solve the expression
          f = utils.timeout(timeout_len)(lambda: sympy.solve(x_eq, x_symbol))
          x_by_y_in_x_eq = f()
          if len(x_by_y_in_x_eq) > 1:
            raise NotImplementedError('Do not support multiple values.')
          x_by_y_in_x_eq = analysis_by_sympy.sympy2str(x_by_y_in_x_eq[0])

          # check
          all_vars = set(eq_x_scope.keys())
          all_vars.update(self.dvar_names + self.dpar_names)
          if utils.contain_unknown_symbol(x_by_y_in_x_eq, all_vars):
            logger.info('\tfailed because contain unknown symbols.')
            results['status'] = 'sympy_failed'
          else:
            logger.info('\tsuccess.')

            # substituted codes
            subs_codes = [f'{expr.var_name} = {expr.code}'
                          for expr in self.x_eq_group.sub_exprs[:-1]]
            subs_codes.append(f'{self.x_var} = {x_by_y_in_x_eq}')

            # compile the function
            func_code = f'def func({",".join(self.dvar_names[1:] + self.dpar_names)}):\n'
            for expr in self.y_eq_group.sub_exprs[:-1]:
              func_code += f'  {expr.var_name} = {expr.code}\n'
            func_code += f'  return {x_by_y_in_x_eq}'
            exec(compile(func_code, '', 'exec'), eq_x_scope)

            # set results
            results['status'] = 'sympy_success'
            results['subs'] = subs_codes
            results['f'] = eq_x_scope['func']
        except NotImplementedError:
          logger.info('\tfailed because the equation is too complex.')
          results['status'] = 'sympy_failed'
        except KeyboardInterrupt:
          logger.info(f'\tfailed because {timeout_len} s timeout.')
          results['status'] = 'sympy_failed'
      else:
        results['status'] = 'escape'
      self.analyzed_results['x_by_y_in_x_eq'] = results
    return self.analyzed_results['x_by_y_in_x_eq']



# -*- coding: utf-8 -*-

import numpy

__all__ = [
  'cholesky', 'cond', 'det', 'eig', 'eigh', 'eigvals', 'eigvalsh', 'inv', 'svd',
  'lstsq', 'matrix_power', 'matrix_rank', 'norm', 'pinv', 'qr',
]

cholesky = numpy.linalg.cholesky
cond = numpy.linalg.cond
det = numpy.linalg.det
eig = numpy.linalg.eig
eigh = numpy.linalg.eigh
eigvals = numpy.linalg.eigvals
eigvalsh = numpy.linalg.eigvalsh
inv = numpy.linalg.inv
svd = numpy.linalg.svd
lstsq = numpy.linalg.lstsq
matrix_power = numpy.linalg.matrix_power
matrix_rank = numpy.linalg.matrix_rank
norm = numpy.linalg.norm
pinv = numpy.linalg.pinv
qr = numpy.linalg.qr



# -*- coding: utf-8 -*-


import abc

from brainpy import math

__all__ = [
  'AbstractDelay',
  'ConstantDelay',
  'VaryingDelay',
  'NeutralDelay',
]


class AbstractDelay(abc.ABC):
  def __setitem__(self, time, value):
    pass

  def __getitem__(self, time):
    pass


class ConstantDelay(AbstractDelay):
  def __init__(self, v0, delay_len, before_t0=0., t0=0., dt=None):
    # size
    self.size = math.shape(v0)

    # delay_len
    self.delay_len = delay_len
    self.dt = math.get_dt() if dt is None else dt
    self.num_delay = int(math.ceil(delay_len / self.dt))

    # other variables
    self._delay_in = self.num_delay - 1
    self._delay_out = 0
    self.current_time = t0

    # before_t0
    self.before_t0 = before_t0

    # delay data
    self.data = math.zeros((self.num_delay + 1,) + self.size)
    if callable(before_t0):
      for i in range(self.num_delay):
        self.data[i] = before_t0(t0 + (i - self.num_delay) * self.dt)
    else:
      self.data[:-1] = before_t0
    self.data[-1] = v0

  def __setitem__(self, time, value):  # push
    self.data[self._delay_in] = value
    self.current_time = time

  def __getitem__(self, time):  # pull
    diff = self.current_time - time
    m = math.ceil(diff / self.dt)
    return self.data[self._delay_out]

  def update(self):
    self._delay_in = (self._delay_in + 1) % self.num_delay
    self._delay_out = (self._delay_out + 1) % self.num_delay


class VaryingDelay(AbstractDelay):
  def __init__(self):
    pass


class NeutralDelay(AbstractDelay):
  def __init__(self):
    pass



# -*- coding: utf-8 -*-


import ast

from brainpy import tools

__all__ = [
  'control_transform',
]


def control_transform(f):
  return f


'''
if V >= self.V_th:
  self.V[i] = self.V_reset
  self.spike[i] = 1.
  self.t_last_spike[i] = _t
  self.refractory[i] = True
else:
  self.spike[i] = 0.
  self.V[i] = V
  self.refractory[i] = False
'''

string = '''
def update(self, _t, _i):
  for i in range(self.num):
    if _t - self.t_last_spike[i] <= self.t_refractory:
      self.refractory[i] = True
    else:
      V = self.int_V(self.V[i], _t, self.input[i])
      if V >= self.V_th:
        self.V[i] = self.V_reset
        self.spike[i] = 1.
        self.t_last_spike[i] = _t
        self.refractory[i] = True
      else:
        self.spike[i] = 0.
        self.V[i] = V
        self.refractory[i] = False
    self.input[i] = 0.
'''


class NodeTransformer(ast.NodeTransformer):
  pass


def _cond_syntax(tree):
  if len(tree.orelse) == 1 and isinstance(tree.orelse[0], ast.If):
    raise ValueError('Cannot analyze multiple conditional syntax with "if .. elif .. ". '
                     'Please code the code with "if .. else .. " syntax.')
  # code for "if" flow

  # code for "else" flow


def _for_loop_syntax(tree):
  pass


def _while_loop(tree):
  pass


def _transform(code):
  tree = ast.parse(code)

  # arguments
  args = tools.ast2code(ast.fix_missing_locations(tree.body[0].args)).split(',')



# -*- coding: utf-8 -*-

import brainpy as bp
import pytest


def test_IJConn():
    conn = bp.connect.IJConn(i=bp.math.array([0, 1, 2]),
                             j=bp.math.array([0, 0, 0]))
    conn = conn(pre_size=5, post_size=3)

    print(conn.requires('pre2post'))
    print(conn.requires(bp.connect.CONN_MAT))


def test_MatConn1():
    conn = bp.connect.MatConn(conn_mat=bp.math.random.randint(2, size=(5, 3), dtype=bp.math.bool_))
    conn = conn(pre_size=5, post_size=3)

    print(conn.requires('pre2post'))
    print(conn.requires(bp.connect.CONN_MAT))


def test_MatConn2():
    conn = bp.connect.MatConn(conn_mat=bp.math.random.randint(2, size=(5, 3), dtype=bp.math.bool_))
    with pytest.raises(AssertionError):
      conn = conn(pre_size=5, post_size=1)



# -*- coding: utf-8 -*-


from pprint import pprint

import brainpy as bp


class GABAa_without_Variable(bp.TwoEndConn):
  def __init__(self, pre, post, conn, delay=0., g_max=0.1, E=-75.,
               alpha=12., beta=0.1, T=1.0, T_duration=1.0, **kwargs):
    super(GABAa_without_Variable, self).__init__(pre=pre, post=post, **kwargs)

    # parameters
    self.g_max = g_max
    self.E = E
    self.alpha = alpha
    self.beta = beta
    self.T = T
    self.T_duration = T_duration
    self.delay = delay

    # connections
    self.conn = conn(pre.size, post.size)
    self.conn_mat = self.conn.requires(bp.connect.CONN_MAT)
    self.size = bp.math.shape(self.conn_mat)

    # variables
    self.t_last_pre_spike = bp.math.ones(self.size) * -1e7
    self.s = bp.math.zeros(self.size)
    self.g = self.register_constant_delay('g', size=self.size, delay=delay)

  @bp.odeint
  def int_s(self, s, t, TT):
    return self.alpha * TT * (1 - s) - self.beta * s

  def update(self, _t, _i):
    spike = bp.math.reshape(self.pre.spikes, (self.pre.num, 1)) * self.conn_mat
    self.t_last_pre_spike[:] = bp.math.where(spike, _t, self.t_last_pre_spike)
    TT = ((_t - self.t_last_pre_spike) < self.T_duration) * self.T
    self.s[:] = self.int_s(self.s, _t, TT)
    self.g.push(self.g_max * self.s)
    g = self.g.pull()
    self.post.inputs -= bp.math.sum(g, axis=0) * (self.post.V - self.E)


class HH_without_Variable(bp.NeuGroup):
  def __init__(self, size, ENa=55., EK=-90., EL=-65, C=1.0,
               gNa=35., gK=9., gL=0.1, V_th=20., phi=5.0, **kwargs):
    super(HH_without_Variable, self).__init__(size=size, **kwargs)

    # parameters
    self.ENa = ENa
    self.EK = EK
    self.EL = EL
    self.C = C
    self.gNa = gNa
    self.gK = gK
    self.gL = gL
    self.V_th = V_th
    self.phi = phi

    # variables
    self.V = bp.math.ones(self.num) * -65.
    self.h = bp.math.ones(self.num) * 0.6
    self.n = bp.math.ones(self.num) * 0.32
    self.inputs = bp.math.zeros(self.num)
    self.spikes = bp.math.zeros(self.num, dtype=bp.math.bool_)

  @bp.odeint
  def integral(self, V, h, n, t, Iext):
    alpha = 0.07 * bp.math.exp(-(V + 58) / 20)
    beta = 1 / (bp.math.exp(-0.1 * (V + 28)) + 1)
    dhdt = self.phi * (alpha * (1 - h) - beta * h)

    alpha = -0.01 * (V + 34) / (bp.math.exp(-0.1 * (V + 34)) - 1)
    beta = 0.125 * bp.math.exp(-(V + 44) / 80)
    dndt = self.phi * (alpha * (1 - n) - beta * n)

    m_alpha = -0.1 * (V + 35) / (bp.math.exp(-0.1 * (V + 35)) - 1)
    m_beta = 4 * bp.math.exp(-(V + 60) / 18)
    m = m_alpha / (m_alpha + m_beta)
    INa = self.gNa * m ** 3 * h * (V - self.ENa)
    IK = self.gK * n ** 4 * (V - self.EK)
    IL = self.gL * (V - self.EL)
    dVdt = (- INa - IK - IL + Iext) / self.C

    return dVdt, dhdt, dndt

  def update(self, _t, _i):
    V, h, n = self.integral(self.V, self.h, self.n, _t, self.inputs)
    self.spikes[:] = bp.math.logical_and(self.V < self.V_th, V >= self.V_th)
    self.V[:] = V
    self.h[:] = h
    self.n[:] = n
    self.inputs[:] = 0.


def test_subset_integrator():
  neu = HH_without_Variable(10, monitors=['spikes', 'V'])
  syn = GABAa_without_Variable(pre=neu, post=neu, conn=bp.connect.All2All(include_self=False))
  syn.g_max = 0.1 / neu.num
  net = bp.Network(neu, syn)

  ints = net.ints()
  print()
  print(ints)

  ode_ints = ints.subset(bp.integrators.ODE_INT)
  print(ode_ints)
  assert len(ode_ints) == 2


def test_neu_vars_1():
  neu = HH_without_Variable(10, monitors=['spikes', 'V'])
  vars = neu.vars()

  print()
  print(vars)
  assert len(vars) == 0


class HH_with_Variable(bp.NeuGroup):
  def __init__(self, size, ENa=55., EK=-90., EL=-65, C=1.0,
               gNa=35., gK=9., gL=0.1, V_th=20., phi=5.0, **kwargs):
    super(HH_with_Variable, self).__init__(size=size, **kwargs)

    # parameters
    self.ENa = ENa
    self.EK = EK
    self.EL = EL
    self.C = C
    self.gNa = gNa
    self.gK = gK
    self.gL = gL
    self.V_th = V_th
    self.phi = phi

    # variables
    self.V = bp.math.Variable(bp.math.ones(self.num) * -65.)
    self.h = bp.math.Variable(bp.math.ones(self.num) * 0.6)
    self.n = bp.math.Variable(bp.math.ones(self.num) * 0.32)
    self.inputs = bp.math.Variable(bp.math.zeros(self.num))
    self.spikes = bp.math.Variable(bp.math.zeros(self.num, dtype=bp.math.bool_))

  @bp.odeint
  def integral(self, V, h, n, t, Iext):
    alpha = 0.07 * bp.math.exp(-(V + 58) / 20)
    beta = 1 / (bp.math.exp(-0.1 * (V + 28)) + 1)
    dhdt = self.phi * (alpha * (1 - h) - beta * h)

    alpha = -0.01 * (V + 34) / (bp.math.exp(-0.1 * (V + 34)) - 1)
    beta = 0.125 * bp.math.exp(-(V + 44) / 80)
    dndt = self.phi * (alpha * (1 - n) - beta * n)

    m_alpha = -0.1 * (V + 35) / (bp.math.exp(-0.1 * (V + 35)) - 1)
    m_beta = 4 * bp.math.exp(-(V + 60) / 18)
    m = m_alpha / (m_alpha + m_beta)
    INa = self.gNa * m ** 3 * h * (V - self.ENa)
    IK = self.gK * n ** 4 * (V - self.EK)
    IL = self.gL * (V - self.EL)
    dVdt = (- INa - IK - IL + Iext) / self.C

    return dVdt, dhdt, dndt

  def update(self, _t, _i):
    V, h, n = self.integral(self.V, self.h, self.n, _t, self.inputs)
    self.spikes[:] = bp.math.logical_and(self.V < self.V_th, V >= self.V_th)
    self.V[:] = V
    self.h[:] = h
    self.n[:] = n
    self.inputs[:] = 0.


def test_neu_vars_2():
  neu = HH_with_Variable(10, monitors=['spikes', 'V'])
  vars = neu.vars()
  print()
  print(vars.keys())

  vars = neu.vars(method='relative')
  print()
  print(vars.keys())


def test_neu_nodes_1():
  neu = HH_with_Variable(10, monitors=['spikes', 'V'])
  print()
  print(neu.nodes().keys())
  assert len(neu.nodes()) == 1

  print()
  print(neu.nodes(method='relative').keys())
  assert len(neu.nodes(method='relative')) == 1


def test_neu_ints_1():
  neu = HH_with_Variable(10, monitors=['spikes', 'V'])
  print()
  print(neu.ints().keys())
  assert len(neu.ints()) == 1

  print()
  print(neu.ints(method='relative').keys())
  assert len(neu.ints(method='relative')) == 1


class GABAa_with_Variable(bp.TwoEndConn):
  def __init__(self, pre, post, conn, delay=0., g_max=0.1, E=-75.,
               alpha=12., beta=0.1, T=1.0, T_duration=1.0, **kwargs):
    super(GABAa_with_Variable, self).__init__(pre=pre, post=post, **kwargs)

    # parameters
    self.g_max = g_max
    self.E = E
    self.alpha = alpha
    self.beta = beta
    self.T = T
    self.T_duration = T_duration
    self.delay = delay

    # connections
    self.conn = conn(pre.size, post.size)
    self.conn_mat = self.conn.requires(bp.connect.CONN_MAT)
    self.size = bp.math.shape(self.conn_mat)

    # variables
    self.t_last_pre_spike = bp.math.Variable(bp.math.ones(self.size) * -1e7)
    self.s = bp.math.Variable(bp.math.zeros(self.size))
    self.g = self.register_constant_delay('g', size=self.size, delay=delay)

  @bp.odeint
  def int_s(self, s, t, TT):
    return self.alpha * TT * (1 - s) - self.beta * s

  def update(self, _t, _i):
    spike = bp.math.reshape(self.pre.spikes, (self.pre.num, 1)) * self.conn_mat
    self.t_last_pre_spike[:] = bp.math.where(spike, _t, self.t_last_pre_spike)
    TT = ((_t - self.t_last_pre_spike) < self.T_duration) * self.T
    self.s[:] = self.int_s(self.s, _t, TT)
    self.g.push(self.g_max * self.s)
    g = self.g.pull()
    self.post.inputs -= bp.math.sum(g, axis=0) * (self.post.V - self.E)


def test_net_1():
  neu = HH_without_Variable(10, monitors=['spikes', 'V'])
  syn = GABAa_without_Variable(pre=neu, post=neu, conn=bp.connect.All2All(include_self=False))
  net = bp.Network(neu=neu, syn=syn)

  # variables
  print()
  pprint(list(net.vars().keys()))
  assert len(net.vars()) == 3

  print()
  pprint(list(net.vars(method='relative').keys()))
  assert len(net.vars(method='relative')) == 3

  # nodes
  print()
  pprint(list(net.nodes().keys()))
  assert len(net.nodes()) == 4

  print()
  pprint(list(net.nodes(method='relative').keys()))
  assert len(net.nodes(method='relative')) == 5

  # ints
  print()
  pprint(list(net.ints().keys()))
  assert len(net.ints()) == 2

  print()
  pprint(list(net.ints(method='relative').keys()))
  assert len(net.ints(method='relative')) == 3


def test_net_vars_2():
  neu = HH_with_Variable(10, monitors=['spikes', 'V'])
  syn = GABAa_with_Variable(pre=neu, post=neu, conn=bp.connect.All2All(include_self=False))
  net = bp.Network(neu=neu, syn=syn)

  # variables
  print()
  pprint(list(net.vars().keys()))
  # assert len(net.vars()) == 3

  print()
  pprint(list(net.vars(method='relative').keys()))
  # assert len(net.vars(method='relative')) == 3

  # nodes
  print()
  pprint(list(net.nodes().keys()))
  assert len(net.nodes()) == 4

  print()
  pprint(list(net.nodes(method='relative').keys()))
  assert len(net.nodes(method='relative')) == 5

  # ints
  print()
  pprint(list(net.ints().keys()))
  assert len(net.ints()) == 2

  print()
  pprint(list(net.ints(method='relative').keys()))
  assert len(net.ints(method='relative')) == 3



# -*- coding: utf-8 -*-

import inspect

from brainpy import errors, math
from brainpy.integrators import constants
from brainpy.integrators.analysis_by_ast import separate_variables
from . import common

__all__ = [
  'euler',
  'heun',
  'milstein',
  'exponential_euler',
]


class Tools(object):

  @staticmethod
  def df_and_dg(code_lines, variables, parameters):
    # 1. df
    # df = f(x, t, *args)
    all_df = [f'{var}_df' for var in variables]
    code_lines.append(f'  {", ".join(all_df)} = f({", ".join(variables + parameters)})')

    # 2. dg
    # dg = g(x, t, *args)
    all_dg = [f'{var}_dg' for var in variables]
    code_lines.append(f'  {", ".join(all_dg)} = g({", ".join(variables + parameters)})')
    code_lines.append('  ')

  @staticmethod
  def dfdt(code_lines, variables, vdt):
    for var in variables:
      code_lines.append(f'  {var}_dfdt = {var}_df * {vdt}')
    code_lines.append('  ')

  @staticmethod
  def noise_terms(code_lines, variables):
    # num_vars = len(variables)
    # if num_vars > 1:
    #     code_lines.append(f'  all_dW = math.normal(0.0, dt_sqrt, ({num_vars},)+math.shape({variables[0]}_dg))')
    #     for i, var in enumerate(variables):
    #         code_lines.append(f'  {var}_dW = all_dW[{i}]')
    # else:
    #     var = variables[0]
    #     code_lines.append(f'  {var}_dW = math.normal(0.0, dt_sqrt, math.shape({var}))')
    # code_lines.append('  ')

    for var in variables:
      code_lines.append(f'  {var}_dW = math.random.normal(0.000, dt_sqrt, math.shape({var}))')
    code_lines.append('  ')


class Wrapper(object):

  @staticmethod
  def wrap(wrapper, f, g, dt, intg_type, var_type, wiener_type, show_code):
    """The base function to format a SRK method.

    Parameters
    ----------
    f : callable
        The drift function of the SDE_INT.
    g : callable
        The diffusion function of the SDE_INT.
    dt : float
        The numerical precision.
    intg_type : str
        "utils.ITO_SDE" : Ito's Stochastic Calculus.
        "utils.STRA_SDE" : Stratonovich's Stochastic Calculus.
    wiener_type : str
    var_type : str
        "scalar" : with the shape of ().
        "population" : with the shape of (N,) or (N1, N2) or (N1, N2, ...).
        "system": with the shape of (d, ), (d, N), or (d, N1, N2).
    show_code : bool
        Whether show the formatted code.

    Returns
    -------
    numerical_func : callable
        The numerical function.
    """

    intg_type = constants.ITO_SDE if intg_type is None else intg_type
    var_type = constants.SCALAR_VAR if var_type is None else var_type
    wiener_type = constants.SCALAR_WIENER if wiener_type is None else wiener_type
    if intg_type not in constants.SUPPORTED_INTG_TYPE:
      raise errors.IntegratorError(f'Currently, BrainPy only support SDE_INT types: '
                                   f'{constants.SUPPORTED_INTG_TYPE}. But we got {intg_type}.')
    if var_type not in constants.SUPPORTED_VAR_TYPE:
      raise errors.IntegratorError(f'Currently, BrainPy only supports variable types: '
                                   f'{constants.SUPPORTED_VAR_TYPE}. But we got {var_type}.')
    if wiener_type not in constants.SUPPORTED_WIENER_TYPE:
      raise errors.IntegratorError(f'Currently, BrainPy only supports Wiener '
                                   f'Process types: {constants.SUPPORTED_WIENER_TYPE}. '
                                   f'But we got {wiener_type}.')

    show_code = False if show_code is None else show_code
    dt = math.get_dt() if dt is None else dt

    if f is not None and g is not None:
      return wrapper(f=f, g=g, dt=dt, show_code=show_code, intg_type=intg_type,
                     var_type=var_type, wiener_type=wiener_type)

    elif f is not None:
      return lambda g: wrapper(f=f, g=g, dt=dt, show_code=show_code, intg_type=intg_type,
                               var_type=var_type, wiener_type=wiener_type)

    elif g is not None:
      return lambda f: wrapper(f=f, g=g, dt=dt, show_code=show_code, intg_type=intg_type,
                               var_type=var_type, wiener_type=wiener_type)

    else:
      raise ValueError('Must provide "f" or "g".')

  @staticmethod
  def exp_euler(f, g, dt, intg_type, var_type, wiener_type, show_code):
    try:
      import sympy
      from brainpy.integrators import analysis_by_sympy
    except ModuleNotFoundError:
      raise errors.PackageMissingError('SymPy must be installed when using exponential euler methods.')

    if var_type == constants.SYSTEM_VAR:
      raise errors.IntegratorError(f'Exponential Euler method do not support {var_type} variable type.')
    if intg_type != constants.ITO_SDE:
      raise errors.IntegratorError(f'Exponential Euler method only supports Ito integral, but we got {intg_type}.')

    vdt, variables, parameters, arguments, func_name = common.basic_info(f=f, g=g)
    arguments = list(arguments) + [f'{vdt}={dt}']

    # 1. code scope
    closure_vars = inspect.getclosurevars(f)
    code_scope = dict(closure_vars.nonlocals)
    code_scope.update(dict(closure_vars.globals))
    code_scope['f'] = f
    code_scope['g'] = g
    code_scope['math'] = math
    # code_scope['exp'] = math.exp

    # 2. code lines
    code_lines = [f'def {func_name}({", ".join(arguments)}):']
    code_lines.append(f'  {vdt}_sqrt = {vdt} ** 0.5')

    # 2.1 dg
    # dg = g(x, t, *args)
    all_dg = [f'{var}_dg' for var in variables]
    code_lines.append(f'  {", ".join(all_dg)} = g({", ".join(variables + parameters)})')
    code_lines.append('  ')

    # 2.2 dW
    Tools.noise_terms(code_lines, variables)

    # 2.3 dgdW
    # ----
    # SCALAR_WIENER : dg * dW
    # VECTOR_WIENER : math.sum(dg * dW, axis=-1)

    if wiener_type == constants.SCALAR_WIENER:
      for var in variables:
        code_lines.append(f'  {var}_dgdW = {var}_dg * {var}_dW')
    else:
      for var in variables:
        code_lines.append(f'  {var}_dgdW = math.sum({var}_dg * {var}_dW, axis=-1)')
    code_lines.append('  ')

    # 2.4 new var
    # ----
    analysis = separate_variables(f)
    variables_for_returns = analysis['variables_for_returns']
    expressions_for_returns = analysis['expressions_for_returns']
    for vi, (key, vars) in enumerate(variables_for_returns.items()):
      # separate variables
      sd_variables = []
      for v in vars:
        if len(v) > 1:
          raise ValueError('Cannot analyze multi-assignment code line.')
        sd_variables.append(v[0])
      expressions = expressions_for_returns[key]
      var_name = variables[vi]
      diff_eq = analysis_by_sympy.SingleDiffEq(var_name=var_name,
                                               variables=sd_variables,
                                               expressions=expressions,
                                               derivative_expr=key,
                                               scope=code_scope,
                                               func_name=func_name)

      f_expressions = diff_eq.get_f_expressions(substitute_vars=diff_eq.var_name)

      # code lines
      code_lines.extend([f"  {str(expr)}" for expr in f_expressions[:-1]])

      # get the linear system using sympy
      f_res = f_expressions[-1]
      df_expr = analysis_by_sympy.str2sympy(f_res.code).expr.expand()
      s_df = sympy.Symbol(f"{f_res.var_name}")
      code_lines.append(f'  {s_df.name} = {analysis_by_sympy.sympy2str(df_expr)}')
      var = sympy.Symbol(diff_eq.var_name, real=True)

      # get df part
      s_linear = sympy.Symbol(f'_{diff_eq.var_name}_linear')
      s_linear_exp = sympy.Symbol(f'_{diff_eq.var_name}_linear_exp')
      s_df_part = sympy.Symbol(f'_{diff_eq.var_name}_df_part')
      if df_expr.has(var):
        # linear
        linear = sympy.collect(df_expr, var, evaluate=False)[var]
        code_lines.append(f'  {s_linear.name} = {analysis_by_sympy.sympy2str(linear)}')
        # linear exponential
        code_lines.append(f'  {s_linear_exp.name} = math.exp({linear.name} * {vdt})')
        # df part
        df_part = (s_linear_exp - 1) / s_linear * s_df
        code_lines.append(f'  {s_df_part.name} = {analysis_by_sympy.sympy2str(df_part)}')

      else:
        # linear exponential
        code_lines.append(f'  {s_linear_exp.name} = {vdt}_sqrt')
        # df part
        code_lines.append(f'  {s_df_part.name} = {s_df.name} * {vdt}')

      # update expression
      update = var + s_df_part

      # The actual update step
      code_lines.append(f'  {diff_eq.var_name}_new = {analysis_by_sympy.sympy2str(update)} + {var_name}_dgdW')
      code_lines.append('')

    # returns
    new_vars = [f'{var}_new' for var in variables]
    code_lines.append(f'  return {", ".join(new_vars)}')

    # return and compile
    return common.compile_and_assign_attrs(
      code_lines=code_lines, code_scope=code_scope, show_code=show_code,
      variables=variables, parameters=parameters, func_name=func_name,
      intg_type=intg_type, var_type=var_type, wiener_type=wiener_type,
      dt=dt, method='exponential_euler', raw_func=dict(f=f, g=g)
    )

  @staticmethod
  def euler_and_heun(f, g, dt, intg_type, var_type, wiener_type, show_code):
    vdt, variables, parameters, arguments, func_name = common.basic_info(f=f, g=g)
    arguments = list(arguments) + [f'{vdt}={dt}']

    # 1. code scope
    code_scope = {'f': f, 'g': g, 'math': math}

    # 2. code lines
    code_lines = [f'def {func_name}({", ".join(arguments)}):']
    code_lines.append(f'  {vdt}_sqrt = {vdt} ** 0.5')

    # 2.1 df, dg
    Tools.df_and_dg(code_lines, variables, parameters)

    # 2.2 dfdt
    Tools.dfdt(code_lines, variables, vdt)

    # 2.3 dW
    Tools.noise_terms(code_lines, variables)

    # 2.3 dgdW
    # ----
    # SCALAR_WIENER : dg * dW
    # VECTOR_WIENER : math.sum(dg * dW, axis=-1)

    if wiener_type == constants.SCALAR_WIENER:
      for var in variables:
        code_lines.append(f'  {var}_dgdW = {var}_dg * {var}_dW')
    else:
      for var in variables:
        code_lines.append(f'  {var}_dgdW = math.sum({var}_dg * {var}_dW, axis=-1)')
    code_lines.append('  ')

    if intg_type == constants.ITO_SDE:
      method = 'euler'
      # 2.4 new var
      # ----
      # y = x + dfdt + dgdW
      for var in variables:
        code_lines.append(f'  {var}_new = {var} + {var}_dfdt + {var}_dgdW')
      code_lines.append('  ')

    elif intg_type == constants.STRA_SDE:
      method = 'heun'

      # 2.4  y_bar = x + math.sum(dgdW, axis=-1)
      all_bar = [f'{var}_bar' for var in variables]
      for var in variables:
        code_lines.append(f'  {var}_bar = {var} + {var}_dgdW')
      code_lines.append('  ')

      # 2.5  dg_bar = g(y_bar, t, *args)
      all_dg_bar = [f'{var}_dg_bar' for var in variables]
      code_lines.append(f'  {", ".join(all_dg_bar)} = g({", ".join(all_bar + parameters)})')

      # 2.6 dgdW2
      # ----
      # SCALAR_WIENER : dgdW2 = dg_bar * dW
      # VECTOR_WIENER : dgdW2 = math.sum(dg_bar * dW, axis=-1)
      if wiener_type == constants.SCALAR_WIENER:
        for var in variables:
          code_lines.append(f'  {var}_dgdW2 = {var}_dg_bar * {var}_dW')
      else:
        for var in variables:
          code_lines.append(f'  {var}_dgdW2 = math.sum({var}_dg_bar * {var}_dW, axis=-1)')
      code_lines.append('  ')

      # 2.7 new var
      # ----
      # y = x + dfdt + 0.5 * (dgdW + dgdW2)
      for var in variables:
        code_lines.append(f'  {var}_new = {var} + {var}_dfdt + 0.5 * ({var}_dgdW + {var}_dgdW2)')
      code_lines.append('  ')
    else:
      raise ValueError(f'Unknown SDE_INT type: {intg_type}. We only '
                       f'supports {constants.SUPPORTED_INTG_TYPE}.')

    # returns
    new_vars = [f'{var}_new' for var in variables]
    code_lines.append(f'  return {", ".join(new_vars)}')

    # return and compile
    return common.compile_and_assign_attrs(
      code_lines=code_lines, code_scope=code_scope, show_code=show_code,
      variables=variables, parameters=parameters, func_name=func_name,
      intg_type=intg_type, var_type=var_type, wiener_type=wiener_type,
      dt=dt, method=method, raw_func=dict(f=f, g=g))

  @staticmethod
  def milstein(f, g, dt, intg_type, var_type, wiener_type, show_code):
    vdt, variables, parameters, arguments, func_name = common.basic_info(f=f, g=g)
    arguments = list(arguments) + [f'{vdt}={dt}']

    # 1. code scope
    code_scope = {'f': f, 'g': g, 'math': math}

    # 2. code lines
    code_lines = [f'def {func_name}({", ".join(arguments)}):']
    code_lines.append(f'  {vdt}_sqrt = {vdt} ** 0.5')

    # 2.1 df, dg
    Tools.df_and_dg(code_lines, variables, parameters)

    # 2.2 dfdt
    Tools.dfdt(code_lines, variables, vdt)

    # 2.3 dW
    Tools.noise_terms(code_lines, variables)

    # 2.3 dgdW
    # ----
    # dg * dW
    for var in variables:
      code_lines.append(f'  {var}_dgdW = {var}_dg * {var}_dW')
    code_lines.append('  ')

    # 2.4  df_bar = x + dfdt + math.sum(dg * dt_sqrt, axis=-1)
    all_df_bar = [f'{var}_df_bar' for var in variables]
    if wiener_type == constants.SCALAR_WIENER:
      for var in variables:
        code_lines.append(f'  {var}_df_bar = {var} + {var}_dfdt + {var}_dg * {vdt}_sqrt')
    else:
      for var in variables:
        code_lines.append(f'  {var}_df_bar = {var} + {var}_dfdt + math.sum('
                          f'{var}_dg * {vdt}_sqrt, axis=-1)')

    # 2.5  dg_bar = g(y_bar, t, *args)
    all_dg_bar = [f'{var}_dg_bar' for var in variables]
    code_lines.append(f'  {", ".join(all_dg_bar)} = g({", ".join(all_df_bar + parameters)})')
    code_lines.append('  ')

    # 2.6 dgdW2
    # ----
    # dgdW2 = 0.5 * (dg_bar - dg) * (dW * dW / dt_sqrt - dt_sqrt)
    if intg_type == constants.ITO_SDE:
      for var in variables:
        code_lines.append(f'  {var}_dgdW2 = 0.5 * ({var}_dg_bar - {var}_dg) * '
                          f'({var}_dW * {var}_dW / {vdt}_sqrt - {vdt}_sqrt)')
    elif intg_type == constants.STRA_SDE:
      for var in variables:
        code_lines.append(f'  {var}_dgdW2 = 0.5 * ({var}_dg_bar - {var}_dg) * '
                          f'{var}_dW * {var}_dW / {vdt}_sqrt')
    else:
      raise ValueError(f'Unknown SDE_INT type: {intg_type}')
    code_lines.append('  ')

    # 2.7 new var
    # ----
    # SCALAR_WIENER : y = x + dfdt + dgdW + dgdW2
    # VECTOR_WIENER : y = x + dfdt + math.sum(dgdW + dgdW2, axis=-1)
    if wiener_type == constants.SCALAR_WIENER:
      for var in variables:
        code_lines.append(f'  {var}_new = {var} + {var}_dfdt + {var}_dgdW + {var}_dgdW2')
    elif wiener_type == constants.VECTOR_WIENER:
      for var in variables:
        code_lines.append(f'  {var}_new = {var} + {var}_dfdt + math.sum({var}_dgdW + {var}_dgdW2, axis=-1)')
    else:
      raise ValueError(f'Unknown Wiener Process : {wiener_type}')
    code_lines.append('  ')

    # returns
    new_vars = [f'{var}_new' for var in variables]
    code_lines.append(f'  return {", ".join(new_vars)}')

    # return and compile
    return common.compile_and_assign_attrs(
      code_lines=code_lines, code_scope=code_scope, show_code=show_code,
      variables=variables, parameters=parameters, func_name=func_name,
      intg_type=intg_type, var_type=var_type, wiener_type=wiener_type,
      dt=dt, method='milstein', raw_func=dict(f=f, g=g))


def euler(f=None, g=None, dt=None, intg_type=None, var_type=None, wiener_type=None, show_code=None):
  return Wrapper.wrap(Wrapper.euler_and_heun, f=f, g=g, dt=dt, intg_type=intg_type,
                      var_type=var_type, wiener_type=wiener_type, show_code=show_code)


def heun(f=None, g=None, dt=None, intg_type=None, var_type=None, wiener_type=None, show_code=None):
  if intg_type != constants.STRA_SDE:
    raise errors.IntegratorError(f'Heun method only supports Stranovich integral of SDEs, '
                                 f'but we got {intg_type} integral.')
  return Wrapper.wrap(Wrapper.euler_and_heun, f=f, g=g, dt=dt, intg_type=intg_type,
                      var_type=var_type, wiener_type=wiener_type, show_code=show_code)


def exponential_euler(f=None, g=None, dt=None, intg_type=None, var_type=None,
                      wiener_type=None, show_code=None):
  r"""First order, explicit exponential Euler method.

  For a SDE_INT equation of the form

  .. math::

      d y=(Ay+ F(y))dt + g(y)dW(t) = f(y)dt + g(y)dW(t), \quad y(0)=y_{0}

  its schema is given by [1]_

  .. math::

      y_{n+1} & =e^{\Delta t A}(y_{n}+ g(y_n)\Delta W_{n})+\varphi(\Delta t A) F(y_{n}) \Delta t \\
       &= y_n + \Delta t \varphi(\Delta t A) f(y) + e^{\Delta t A}g(y_n)\Delta W_{n}

  where :math:`\varphi(z)=\frac{e^{z}-1}{z}`.

  Parameters
  ----------
  diff_eq : DiffEquation
      The differential equation.

  Returns
  -------
  func : callable
      The one-step numerical integrator function.

  References
  ----------
  .. [1] Erdoğan, Utku, and Gabriel J. Lord. "A new class of exponential integrators for stochastic
         differential equations with multiplicative noise." arXiv preprint arXiv:1608.07096 (2016).
  """

  return Wrapper.wrap(Wrapper.exp_euler, f=f, g=g, dt=dt, intg_type=intg_type,
                      var_type=var_type, wiener_type=wiener_type, show_code=show_code)


def milstein(f=None, g=None, dt=None, intg_type=None, var_type=None, wiener_type=None, show_code=None):
  return Wrapper.wrap(Wrapper.milstein, f=f, g=g, dt=dt, intg_type=intg_type,
                      var_type=var_type, wiener_type=wiener_type, show_code=show_code)
